from collections import OrderedDict

import pandas as pd
from sqlalchemy import create_engine
from clickhouse_sqlalchemy.drivers.base import ClickHouseDialect
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class ClickHouseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the ClickHouse statements.
    """

    name = 'clickhouse'

    def __init__(self, name, connection_data, **kwargs):
        super().__init__(name)
        self.dialect = 'clickhouse'
        self.connection_data = connection_data
        self.renderer = SqlalchemyRender(ClickHouseDialect)
        self.is_connected = False
        self.protocol = connection_data.get('protocol', 'clickhouse')

        # region added for back-compatibility with connections creatad before 11.05.2023
        protocols_map = {
            'native': 'clickhouse+native',
            'http': 'clickhouse+http',
            'https': 'clickhouse+https',
        }
        if self.protocol in protocols_map:
            self.protocol = protocols_map[self.protocol]
        # endregion

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a ClickHouse
        """
        if self.is_connected is True:
            return self.connection

        protocol = self.protocol
        host = self.connection_data['host']
        port = self.connection_data['port']
        user = self.connection_data['user']
        password = self.connection_data['password']
        database = self.connection_data['database']
        url = f'{protocol}://{user}:{password}@{host}:{port}/{database}'
        if self.protocol == 'clickhouse+https':
            url = url + "?protocol=https"

        engine = create_engine(url)
        connection = engine.raw_connection()
        self.is_connected = True
        self.connection = connection
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the ClickHouse database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            cur = connection.cursor()
            try:
                cur.execute('select 1;')
            finally:
                cur.close()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to ClickHouse {self.connection_data["database"]}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in ClickHouse
        :return: returns the records from the current recordset
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        cur = connection.cursor()
        try:
            cur.execute(query)
            result = cur.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    pd.DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            connection.commit()
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
            connection.rollback()
        finally:
            cur.close()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in ClickHouse db
        """
        q = f"SHOW TABLES FROM {self.connection_data['database']}"
        result = self.native_query(q)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"DESCRIBE {table_name}"
        result = self.native_query(q)
        return result


connection_args = OrderedDict(
    protocol={
        'type': ARG_TYPE.STR,
        'protocol': 'The protocol to query clickhouse. Supported: clickhouse, clickhouse+native, clickhouse+http, clickhouse+https. Default: clickhouse'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the ClickHouse server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the ClickHouse server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the ClickHouse server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the ClickHouse server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the ClickHouse server. Must be an integer.'
    }
)

connection_args_example = OrderedDict(
    protocol='clickhouse',
    host='127.0.0.1',
    port=9000,
    user='root',
    password='password',
    database='database'
)



import os

from .session_controller import (
        SessionController,
        ServerSessionContorller,
        )
if os.environ.get("MINDSDB_EXECUTOR_SERVICE_HOST") and os.environ.get("MINDSDB_EXECUTOR_SERVICE_PORT"):
    SessionController = ServerSessionContorller



import os
import dill
import pickle
from typing import Dict
from hashlib import md5

import redis
import sqlite3


class KVStorageHandler:
    """ 
    Simple key-value store. Instances of this handler shall store any information required by other handlers.
    Context should store anything relevant to the storage handler, e.g. CompanyID, UserID, parent handler name, among others. 
    """  # noqa
    def __init__(self, context: Dict, config=None):
        self.config = config if config else os.getenv('MDB_STORAGE_HANDLER_CONFIG')
        self.serializer = pickle if config.get('serializer', '') == 'pickle' else dill
        self.context = self.serializer.dumps(context)  # store serialized

    def _get_context_key(self, key: str):
        serialized_key = self.serializer.dumps(key)
        return md5(serialized_key).hexdigest() + md5(self.context).hexdigest()

    def get(self, key, default_value=None):
        serialized_value = self._get(self._get_context_key(key))
        if serialized_value:
            return self.serializer.loads(serialized_value)
        elif default_value is not None:
            return default_value
        else:
            raise KeyError(f"Key not found: {key}")

    def set(self, key: str, value: object):
        serialized_value = self.serializer.dumps(value)
        self._set(self._get_context_key(key), serialized_value)

    def _get(self, serialized_key):
        raise NotImplementedError()

    def _set(self, serialized_key, serialized_value):
        raise NotImplementedError()


class SqliteStorageHandler(KVStorageHandler):
    """ StorageHandler that uses SQLite as backend. """  # noqa
    def __init__(self, context: Dict, config=None):
        super().__init__(context, config)
        name = self.config["name"] if self.config["name"][-3:] == '.db' else self.config["name"] + '.db'
        path = os.path.join(self.config.get("path", "./"), name)
        self.connection = sqlite3.connect(path)
        self._setup_connection()

    def _setup_connection(self):
        """ Checks that a key-value table exists, otherwise creates it. """  # noqa
        cur = self.connection.cursor()
        if ('store',) not in list(cur.execute("SELECT name FROM sqlite_master WHERE type='table';")):
            cur.execute("""create table store (key text PRIMARY KEY, value text)""")
            self.connection.commit()

    def _get(self, serialized_key):
        cur = self.connection.cursor()
        results = list(cur.execute(f"""select value from store where key='{serialized_key}'"""))
        if results:
            return results[0][0]  # should always be a single match, hence the [0]s
        else:
            return None

    def _set(self, serialized_key, serialized_value):
        cur = self.connection.cursor()
        cur.execute("insert or replace into store values (?, ?)", (serialized_key, serialized_value))
        self.connection.commit()


class RedisStorageHandler(KVStorageHandler):
    """ StorageHandler that uses Redis as backend. """  # noqa
    def __init__(self, context: Dict, config=None):
        super().__init__(context, config)
        assert self.config.get('host', False)
        assert self.config.get('port', False)

        self.connection = redis.Redis(host=self.config['host'], port=self.config['port'])

    def _get(self, serialized_key):
        return self.connection.get(serialized_key)

    def _set(self, serialized_key, serialized_value):
        self.connection.set(serialized_key, serialized_value)



from typing import Optional, Dict

import dill
import pandas as pd
from type_infer.infer import infer_types
import autosklearn.classification as automl_classification
import autosklearn.regression as automl_regression

from .config import ClassificationConfig, RegressionConfig

from mindsdb.integrations.libs.base import BaseMLEngine


class AutoSklearnHandler(BaseMLEngine):
    """
    Integration with the Auto-Sklearn ML library.
    """

    name = 'autosklearn'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        config_args = args['using']

        target_dtype = infer_types(df, 0).to_dict()["dtypes"][target]

        if target_dtype in ['binary', 'categorical', 'tags']:
            config = ClassificationConfig(**config_args)

            model = automl_classification.AutoSklearnClassifier(**vars(config))

        elif target_dtype in ['integer', 'float', 'quantity']:
            config = RegressionConfig(**config_args)

            model = automl_regression.AutoSklearnRegressor(**vars(config))

        else:
            raise Exception('This task is not supported!')

        model.fit(df.drop(target, axis=1), df[target])

        self.model_storage.file_set('model', dill.dumps(model))
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        model = dill.loads(self.model_storage.file_get('model'))
        df = df.drop('__mindsdb_row_id', axis=1)

        predictions = model.predict(df)

        args = self.model_storage.json_get('args')
        df[args['target']] = predictions

        return df


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import mindsdb.microservices_grpc.db.common_pb2 as common__pb2
import mindsdb.microservices_grpc.db.db_pb2 as db__pb2


class DBServiceStub(object):
    """// Represents server ouput without
    // data. like 'OK' or 'ERROR'
    message StatusResponse {
    bool success = 1;
    string error_message = 2;
    }

    // Represents server output with
    // some data inside. SELECT result for instance
    message Response {
    string type = 1;
    bytes data_frame = 2;
    int32 query = 3;
    int32 error_code = 4;
    string error_message = 5;
    }

    Interface exported by the server.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Connect = channel.unary_unary(
                '/DBService/Connect',
                request_serializer=db__pb2.HandlerContext.SerializeToString,
                response_deserializer=common__pb2.StatusResponse.FromString,
                )
        self.CheckConnection = channel.unary_unary(
                '/DBService/CheckConnection',
                request_serializer=db__pb2.HandlerContext.SerializeToString,
                response_deserializer=common__pb2.StatusResponse.FromString,
                )
        self.Disconnect = channel.unary_unary(
                '/DBService/Disconnect',
                request_serializer=db__pb2.HandlerContext.SerializeToString,
                response_deserializer=common__pb2.StatusResponse.FromString,
                )
        self.NativeQuery = channel.unary_unary(
                '/DBService/NativeQuery',
                request_serializer=db__pb2.NativeQueryContext.SerializeToString,
                response_deserializer=common__pb2.Response.FromString,
                )
        self.BinaryQuery = channel.unary_unary(
                '/DBService/BinaryQuery',
                request_serializer=db__pb2.BinaryQueryContext.SerializeToString,
                response_deserializer=common__pb2.Response.FromString,
                )
        self.GetTables = channel.unary_unary(
                '/DBService/GetTables',
                request_serializer=db__pb2.HandlerContext.SerializeToString,
                response_deserializer=common__pb2.Response.FromString,
                )
        self.GetColumns = channel.unary_unary(
                '/DBService/GetColumns',
                request_serializer=db__pb2.ColumnsContext.SerializeToString,
                response_deserializer=common__pb2.Response.FromString,
                )


class DBServiceServicer(object):
    """// Represents server ouput without
    // data. like 'OK' or 'ERROR'
    message StatusResponse {
    bool success = 1;
    string error_message = 2;
    }

    // Represents server output with
    // some data inside. SELECT result for instance
    message Response {
    string type = 1;
    bytes data_frame = 2;
    int32 query = 3;
    int32 error_code = 4;
    string error_message = 5;
    }

    Interface exported by the server.
    """

    def Connect(self, request, context):
        """A simple RPC.

        Establish connection to the specified database

        by creating an appropriate type of Handler instance
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckConnection(self, request, context):
        """A simple RPC.

        Check connection to the specified database

        by creating an appropriate type of Handler instance
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Disconnect(self, request, context):
        """A simple RPC.

        Drop the connection to the specified database

        by creating an appropriate type of Handler instance
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NativeQuery(self, request, context):
        """A simple RPC.

        Execute native query (string) and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BinaryQuery(self, request, context):
        """A simple RPC.

        Execute query (object) and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTables(self, request, context):
        """A simple RPC.

        Performs 'get_tables' request and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetColumns(self, request, context):
        """A simple RPC.

        Performs 'get_columns' request and returns the result

        wrapped into Response object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DBServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Connect': grpc.unary_unary_rpc_method_handler(
                    servicer.Connect,
                    request_deserializer=db__pb2.HandlerContext.FromString,
                    response_serializer=common__pb2.StatusResponse.SerializeToString,
            ),
            'CheckConnection': grpc.unary_unary_rpc_method_handler(
                    servicer.CheckConnection,
                    request_deserializer=db__pb2.HandlerContext.FromString,
                    response_serializer=common__pb2.StatusResponse.SerializeToString,
            ),
            'Disconnect': grpc.unary_unary_rpc_method_handler(
                    servicer.Disconnect,
                    request_deserializer=db__pb2.HandlerContext.FromString,
                    response_serializer=common__pb2.StatusResponse.SerializeToString,
            ),
            'NativeQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.NativeQuery,
                    request_deserializer=db__pb2.NativeQueryContext.FromString,
                    response_serializer=common__pb2.Response.SerializeToString,
            ),
            'BinaryQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.BinaryQuery,
                    request_deserializer=db__pb2.BinaryQueryContext.FromString,
                    response_serializer=common__pb2.Response.SerializeToString,
            ),
            'GetTables': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTables,
                    request_deserializer=db__pb2.HandlerContext.FromString,
                    response_serializer=common__pb2.Response.SerializeToString,
            ),
            'GetColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.GetColumns,
                    request_deserializer=db__pb2.ColumnsContext.FromString,
                    response_serializer=common__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'DBService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DBService(object):
    """// Represents server ouput without
    // data. like 'OK' or 'ERROR'
    message StatusResponse {
    bool success = 1;
    string error_message = 2;
    }

    // Represents server output with
    // some data inside. SELECT result for instance
    message Response {
    string type = 1;
    bytes data_frame = 2;
    int32 query = 3;
    int32 error_code = 4;
    string error_message = 5;
    }

    Interface exported by the server.
    """

    @staticmethod
    def Connect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/Connect',
            db__pb2.HandlerContext.SerializeToString,
            common__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckConnection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/CheckConnection',
            db__pb2.HandlerContext.SerializeToString,
            common__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Disconnect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/Disconnect',
            db__pb2.HandlerContext.SerializeToString,
            common__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def NativeQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/NativeQuery',
            db__pb2.NativeQueryContext.SerializeToString,
            common__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BinaryQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/BinaryQuery',
            db__pb2.BinaryQueryContext.SerializeToString,
            common__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTables(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/GetTables',
            db__pb2.HandlerContext.SerializeToString,
            common__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DBService/GetColumns',
            db__pb2.ColumnsContext.SerializeToString,
            common__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63ommon.proto\"8\n\x0eStatusResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"f\n\x08Response\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x12\n\ndata_frame\x18\x02 \x01(\x0c\x12\r\n\x05query\x18\x03 \x01(\x05\x12\x12\n\nerror_code\x18\x04 \x01(\x05\x12\x15\n\rerror_message\x18\x05 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUSRESPONSE._serialized_start=16
  _STATUSRESPONSE._serialized_end=72
  _RESPONSE._serialized_start=74
  _RESPONSE._serialized_end=176
# @@protoc_insertion_point(module_scope)



import os
import sys
import time
import json
import atexit
import signal
import psutil
import asyncio
import secrets
import traceback
import threading
from packaging import version

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.start import start as start_http
from mindsdb.api.mysql.start import start as start_mysql
from mindsdb.api.mongo.start import start as start_mongo
from mindsdb.api.postgres.start import start as start_postgres
from mindsdb.interfaces.chatbot.chatbot_monitor import start as start_chatbot
from mindsdb.interfaces.jobs.scheduler import start as start_scheduler
from mindsdb.utilities.config import Config
from mindsdb.utilities.ps import is_pid_listen_port, get_child_pids
from mindsdb.utilities.functions import args_parse, get_versions_where_predictors_become_obsolete
from mindsdb.utilities import log
from mindsdb.interfaces.database.integrations import integration_controller
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.utilities.install import install_dependencies
from mindsdb.utilities.fs import create_dirs_recursive, clean_process_marks, clean_unlinked_process_marks
from mindsdb.utilities.telemetry import telemetry_file_exists, disable_telemetry
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.auth import register_oauth_client, get_aws_meta_data

import torch.multiprocessing as mp
try:
    mp.set_start_method('spawn')
except RuntimeError:
    log.logger.info('Torch multiprocessing context already set, ignoring...')


_stop_event = threading.Event()


def close_api_gracefully(apis):
    _stop_event.set()
    try:
        for api in apis.values():
            process = api['process']
            childs = get_child_pids(process.pid)
            for p in childs:
                try:
                    os.kill(p, signal.SIGTERM)
                except Exception:
                    p.kill()
            sys.stdout.flush()
            process.terminate()
            process.join()
            sys.stdout.flush()
    except KeyboardInterrupt:
        sys.exit(0)
    except psutil.NoSuchProcess:
        pass


def do_clean_process_marks():
    while _stop_event.wait(timeout=5) is False:
        clean_unlinked_process_marks()


if __name__ == '__main__':
    # ----------------  __init__.py section ------------------
    clean_process_marks()
    ctx.set_default()
    args = args_parse()

    # ---- CHECK SYSTEM ----
    if not (sys.version_info[0] >= 3 and sys.version_info[1] >= 6):
        print("""
     MindsDB server requires Python >= 3.7 to run

     Once you have Python 3.7 installed you can tun mindsdb as follows:

     1. create and activate venv:
     python3.7 -m venv venv
     source venv/bin/activate

     2. install MindsDB:
     pip3 install mindsdb

     3. Run MindsDB
     python3.7 -m mindsdb

     More instructions in https://docs.mindsdb.com
         """)
        exit(1)

    # --- VERSION MODE ----
    if args is not None and args.version:
        print(f'MindsDB {mindsdb_version}')
        sys.exit(0)

    # --- MODULE OR LIBRARY IMPORT MODE ----
    if args is not None and args.config is not None:
        config_path = args.config
        with open(config_path, 'r') as fp:
            user_config = json.load(fp)
    else:
        user_config = {}
        config_path = 'absent'
    os.environ['MINDSDB_CONFIG_PATH'] = config_path

    mindsdb_config = Config()
    create_dirs_recursive(mindsdb_config['paths'])

    if telemetry_file_exists(mindsdb_config['storage_dir']):
        os.environ['CHECK_FOR_UPDATES'] = '0'
        print('\n x telemetry disabled! \n')
    elif os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false', 'False'] or mindsdb_config.get('cloud', False):
        disable_telemetry(mindsdb_config['storage_dir'])
        print('\n x telemetry disabled \n')
    else:
        print('\n ✓ telemetry enabled \n')

    if os.environ.get('FLASK_SECRET_KEY') is None:
        os.environ['FLASK_SECRET_KEY'] = secrets.token_hex(32)

    # -------------------------------------------------------

    # initialization
    db.init()
    log.initialize_log()

    mp.freeze_support()
    config = Config()

    environment = config.get('environment')
    if environment == 'aws_marketplace':
        try:
            register_oauth_client()
        except Exception as e:
            print(f'Something went wrong during client register: {e}')
    elif environment != 'local':
        try:
            aws_meta_data = get_aws_meta_data()
            Config().update({
                'aws_meta_data': aws_meta_data
            })
        except Exception:
            pass

    is_cloud = config.get('cloud', False)
    # need configure migration behavior by env_variables
    # leave 'is_cloud' for now, but needs to be removed further
    run_migration_separately = os.environ.get("SEPARATE_MIGRATIONS", False)
    if run_migration_separately in (False, "false", "False", 0, "0", ""):
        run_migration_separately = False
    else:
        run_migration_separately = True

    if not is_cloud and not run_migration_separately:
        print('Applying database migrations:')
        try:
            from mindsdb.migrations import migrate
            migrate.migrate_to_head()
        except Exception as e:
            print(f'Error! Something went wrong during DB migrations: {e}')

    if args.verbose is True:
        # Figure this one out later
        pass

    if args.install_handlers is not None:
        handlers_list = [s.strip() for s in args.install_handlers.split(',')]
        # import_meta = handler_meta.get('import', {})
        for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
            if handler_name not in handlers_list:
                continue
            import_meta = handler_meta.get('import', {})
            if import_meta.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - already installed")
                continue
            result = install_dependencies(import_meta.get('dependencies', []))
            if result.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - successfully installed")
            else:
                print(f"{'{0: <18}'.format(handler_name)} - error during dependencies installation: {result.get('error_message', 'unknown error')}")
        sys.exit(0)

    print(f'Version {mindsdb_version}')
    print(f'Configuration file:\n   {config.config_path}')
    print(f"Storage path:\n   {config['paths']['root']}")

    # @TODO Backwards compatibility for tests, remove later
    for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
        import_meta = handler_meta.get('import', {})
        dependencies = import_meta.get('dependencies')
        if import_meta.get('success', False) is not True:
            print(f"Dependencies for the handler '{handler_name}' are not installed by default.\n",
                  f'If you want to use "{handler_name}" please install "{dependencies}"')

    # from mindsdb.utilities.fs import get_marked_processes_and_threads
    # marks = get_marked_processes_and_threads()

    if not is_cloud:
        # region creating permanent integrations
        for integration_name, handler in integration_controller.get_handlers_import_status().items():
            if handler.get('permanent'):
                integration_meta = integration_controller.get(name=integration_name)
                if integration_meta is None:
                    integration_record = db.Integration(
                        name=integration_name,
                        data={},
                        engine=integration_name,
                        company_id=None
                    )
                    db.session.add(integration_record)
                    db.session.commit()
        # endregion

        # region Mark old predictors as outdated
        is_modified = False
        predictor_records = db.session.query(db.Predictor).filter(db.Predictor.deleted_at.is_(None)).all()
        if len(predictor_records) > 0:
            sucess, compatible_versions = get_versions_where_predictors_become_obsolete()
            if sucess is True:
                compatible_versions = [version.parse(x) for x in compatible_versions]
                mindsdb_version_parsed = version.parse(mindsdb_version)
                compatible_versions = [x for x in compatible_versions if x <= mindsdb_version_parsed]
                if len(compatible_versions) > 0:
                    last_compatible_version = compatible_versions[-1]
                    for predictor_record in predictor_records:
                        if (
                            isinstance(predictor_record.mindsdb_version, str)
                            and version.parse(predictor_record.mindsdb_version) < last_compatible_version
                        ):
                            predictor_record.update_status = 'available'
                            is_modified = True
        if is_modified is True:
            db.session.commit()
        # endregion

        for integration_name in config.get('integrations', {}):
            try:
                it = integration_controller.get(integration_name)
                if it is not None:
                    integration_controller.delete(integration_name)
                print(f'Adding: {integration_name}')
                integration_data = config['integrations'][integration_name]
                engine = integration_data.get('type')
                if engine is not None:
                    del integration_data['type']
                integration_controller.add(integration_name, engine, integration_data)
            except Exception as e:
                log.logger.error(f'\n\nError: {e} adding database integration {integration_name}\n\n')

    # @TODO Backwards compatibility for tests, remove later

    if args.api is None:
        api_arr = ['http', 'mysql']
    else:
        api_arr = args.api.split(',')

    with_nlp = False
    if 'nlp' in api_arr:
        with_nlp = True
        api_arr.remove('nlp')

    apis = {
        api: {
            'port': config['api'][api]['port'],
            'process': None,
            'started': False
        } for api in api_arr
    }

    start_functions = {
        'http': start_http,
        'mysql': start_mysql,
        'mongodb': start_mongo,
        'postgres': start_postgres,
        'jobs': start_scheduler,
        'chatbot': start_chatbot
    }

    if config.get('jobs', {}).get('disable') is not True:
        apis['jobs'] = {
            'process': None,
            'started': False
        }

    # disabled on cloud
    if config.get('chatbot', {}).get('disable') is not True and not is_cloud:
        apis['chatbot'] = {
            'process': None,
            'started': False
        }

    ctx = mp.get_context('spawn')
    for api_name, api_data in apis.items():
        if api_data['started']:
            continue
        print(f'{api_name} API: starting...')
        try:
            if api_name == 'http':
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose, args.no_studio, with_nlp))
            else:
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose,))
            p.start()
            api_data['process'] = p
        except Exception as e:
            log.logger.error(f'Failed to start {api_name} API with exception {e}\n{traceback.format_exc()}')
            close_api_gracefully(apis)
            raise e

    atexit.register(close_api_gracefully, apis=apis)

    async def wait_api_start(api_name, pid, port):
        timeout = 60
        start_time = time.time()
        started = is_pid_listen_port(pid, port)
        while (time.time() - start_time) < timeout and started is False:
            await asyncio.sleep(0.5)
            started = is_pid_listen_port(pid, port)
        return api_name, port, started

    async def wait_apis_start():
        futures = [
            wait_api_start(api_name, api_data['process'].pid, api_data['port'])
            for api_name, api_data in apis.items() if 'port' in api_data
        ]
        for i, future in enumerate(asyncio.as_completed(futures)):
            api_name, port, started = await future
            if started:
                print(f"{api_name} API: started on {port}")
            else:
                log.logger.error(f"ERROR: {api_name} API cant start on {port}")

    async def join_process(process, name):
        try:
            process.join()
        except KeyboardInterrupt:
            print('Got keyboard interrupt, stopping APIs')
            close_api_gracefully(apis)
        finally:
            print(f'{name} API: stopped')

    async def gather_apis():
        await asyncio.gather(
            *[join_process(api_data['process'], api_name) for api_name, api_data in apis.items()],
            return_exceptions=False
        )

    ioloop = asyncio.new_event_loop()
    ioloop.run_until_complete(wait_apis_start())

    threading.Thread(target=do_clean_process_marks).start()

    ioloop.run_until_complete(gather_apis())
    ioloop.close()



import os
import shutil
import hashlib
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Union, Optional
from dataclasses import dataclass

from checksumdir import dirhash
try:
    import boto3
except Exception:
    # Only required for remote storage on s3
    pass

from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx


@dataclass(frozen=True)
class RESOURCE_GROUP:
    PREDICTOR = 'predictor'
    INTEGRATION = 'integration'
    TAB = 'tab'


RESOURCE_GROUP = RESOURCE_GROUP()


def copy(src, dst):
    if os.path.isdir(src):
        if os.path.exists(dst):
            if dirhash(src) == dirhash(dst):
                return
        shutil.rmtree(dst, ignore_errors=True)
        shutil.copytree(src, dst)
    else:
        if os.path.exists(dst):
            if hashlib.md5(open(src, 'rb').read()).hexdigest() == hashlib.md5(open(dst, 'rb').read()).hexdigest():
                return
        try:
            os.remove(dst)
        except Exception:
            pass
        shutil.copy2(src, dst)


class BaseFSStore(ABC):
    """Base class for file storage
    """

    def __init__(self):
        self.config = Config()
        self.storage = self.config['paths']['storage']

    @abstractmethod
    def get(self, local_name, base_dir):
        """Copy file/folder from storage to {base_dir}

        Args:
            local_name (str): name of resource (file/folder)
            base_dir (str): path to copy the resource
        """
        pass

    @abstractmethod
    def put(self, local_name, base_dir):
        """Copy file/folder from {base_dir} to storage

        Args:
            local_name (str): name of resource (file/folder)
            base_dir (str): path to folder with the resource
        """
        pass

    @abstractmethod
    def delete(self, remote_name):
        """Delete file/folder from storage

        Args:
            remote_name (str): name of resource
        """
        pass


class LocalFSStore(BaseFSStore):
    """Storage that stores files locally
    """

    def __init__(self):
        super().__init__()

    def get(self, local_name, base_dir):
        remote_name = local_name
        src = os.path.join(self.storage, remote_name)
        dest = os.path.join(base_dir, local_name)
        if not os.path.exists(dest) or os.path.getsize(src) != os.path.getsize(dest):
            copy(src, dest)

    def put(self, local_name, base_dir):
        remote_name = local_name
        copy(
            os.path.join(base_dir, local_name),
            os.path.join(self.storage, remote_name)
        )

    def delete(self, remote_name):
        pass


class S3FSStore(BaseFSStore):
    """Storage that stores files in amazon s3
    """

    def __init__(self):
        super().__init__()
        if 's3_credentials' in self.config['permanent_storage']:
            self.s3 = boto3.client('s3', **self.config['permanent_storage']['s3_credentials'])
        else:
            self.s3 = boto3.client('s3')
        self.bucket = self.config['permanent_storage']['bucket']

    def get(self, local_name, base_dir):
        remote_name = local_name
        remote_ziped_name = f'{remote_name}.tar.gz'
        local_ziped_name = f'{local_name}.tar.gz'
        local_ziped_path = os.path.join(base_dir, local_ziped_name)
        os.makedirs(base_dir, exist_ok=True)
        self.s3.download_file(self.bucket, remote_ziped_name, local_ziped_path)
        shutil.unpack_archive(local_ziped_path, base_dir)
        os.system(f'chmod -R 777 {base_dir}')
        os.remove(local_ziped_path)

    def put(self, local_name, base_dir):
        # NOTE: This `make_archive` function is implemente poorly and will create an empty archive file even if
        # the file/dir to be archived doesn't exist or for some other reason can't be archived
        remote_name = local_name
        shutil.make_archive(
            os.path.join(base_dir, remote_name),
            'gztar',
            root_dir=base_dir,
            base_dir=local_name
        )
        self.s3.upload_file(
            os.path.join(base_dir, f'{remote_name}.tar.gz'),
            self.bucket,
            f'{remote_name}.tar.gz'
        )
        os.remove(os.path.join(base_dir, remote_name + '.tar.gz'))

    def delete(self, remote_name):
        self.s3.delete_object(Bucket=self.bucket, Key=remote_name)


def FsStore():
    storage_location = Config()['permanent_storage']['location']
    if storage_location == 'local':
        return LocalFSStore()
    elif storage_location == 's3':
        return S3FSStore()
    else:
        raise Exception(f"Location: '{storage_location}' not supported")


class FileStorage:
    def __init__(self, resource_group: str, resource_id: int,
                 root_dir: str = 'content', sync: bool = True):
        """
            Args:
                resource_group (str)
                resource_id (int)
                root_dir (str)
                sync (bool)
        """

        self.resource_group = resource_group
        self.resource_id = resource_id
        self.root_dir = root_dir
        self.sync = sync

        self.folder_name = f'{resource_group}_{ctx.company_id}_{resource_id}'

        config = Config()
        self.fs_store = FsStore()
        self.content_path = Path(config['paths'][root_dir])
        self.resource_group_path = self.content_path / resource_group
        self.folder_path = self.resource_group_path / self.folder_name
        if self.folder_path.exists() is False:
            self.folder_path.mkdir(parents=True, exist_ok=True)

    def push(self):
        self.fs_store.put(str(self.folder_name), str(self.resource_group_path))

    def push_path(self, path):
        self.fs_store.put(os.path.join(self.folder_name, path), str(self.resource_group_path))

    def pull(self):
        try:
            self.fs_store.get(str(self.folder_name), str(self.resource_group_path))
        except Exception:
            pass

    def pull_path(self, path, update=True):
        if update is False:
            # not pull from source if object is exists
            if os.path.exists(self.resource_group_path / self.folder_name / path):
                return
        try:
            # TODO not sync if not changed?
            self.fs_store.get(os.path.join(self.folder_name, path), str(self.resource_group_path))
        except Exception:
            pass

    def file_set(self, name, content):
        if self.sync is True:
            self.pull()

        dest_abs_path = self.folder_path / name

        with open(dest_abs_path, 'wb') as fd:
            fd.write(content)

        if self.sync is True:
            self.push()

    def file_get(self, name):

        if self.sync is True:
            self.pull()

        dest_abs_path = self.folder_path / name

        with open(dest_abs_path, 'rb') as fd:
            return fd.read()

    def add(self, path: Union[str, Path], dest_rel_path: Optional[Union[str, Path]] = None):
        """Copy file/folder to persist storage

        Examples:
            Copy file 'args.json' to '{storage}/args.json'
            >>> fs.add('/path/args.json')

            Copy file 'args.json' to '{storage}/folder/opts.json'
            >>> fs.add('/path/args.json', 'folder/opts.json')

            Copy folder 'folder' to '{storage}/folder'
            >>> fs.add('/path/folder')

            Copy folder 'folder' to '{storage}/path/folder'
            >>> fs.add('/path/folder', 'path/folder')

        Args:
            path (Union[str, Path]): path to the resource
            dest_rel_path (Optional[Union[str, Path]]): relative path in storage to file or folder
        """
        if self.sync is True:
            self.pull()

        path = Path(path)
        if isinstance(dest_rel_path, str):
            dest_rel_path = Path(dest_rel_path)

        if dest_rel_path is None:
            dest_abs_path = self.folder_path / path.name
        else:
            dest_abs_path = self.folder_path / dest_rel_path

        copy(
            str(path),
            str(dest_abs_path)
        )

        if self.sync is True:
            self.push()

    def get_path(self, relative_path: Union[str, Path]) -> Path:
        """ Return path to file or folder

        Examples:
            get path to 'opts.json':
            >>> fs.get_path('folder/opts.json')
            ... /path/{storage}/folder/opts.json

        Args:
            relative_path (Union[str, Path]): Path relative to the storage folder

        Returns:
            Path: path to requested file or folder
        """
        if self.sync is True:
            self.pull()

        if isinstance(relative_path, str):
            relative_path = Path(relative_path)
        # relative_path = relative_path.resolve()

        if relative_path.is_absolute():
            raise TypeError('FSStorage.get_path() got absolute path as argument')

        ret_path = self.folder_path / relative_path
        if not ret_path.exists():
            # raise Exception('Path does not exists')
            os.makedirs(ret_path)

        return ret_path

    def delete(self, relative_path: Union[str, Path] = '.') -> Path:
        if isinstance(relative_path, str):
            relative_path = Path(relative_path)

        if relative_path.is_absolute():
            raise TypeError('FSStorage.delete() got absolute path as argument')

        path = (self.folder_path / relative_path).resolve()

        if path == self.folder_path.resolve():
            return self.complete_removal()

        if self.sync is True:
            self.pull()

        if path.exists() is False:
            raise Exception('Path does not exists')

        if path.is_file():
            path.unlink()
        else:
            path.rmdir()

        if self.sync is True:
            self.push()

    def complete_removal(self):
        shutil.rmtree(str(self.folder_path))
        self.fs_store.delete(self.folder_name)


class FileStorageFactory:
    def __init__(self, resource_group: str,
                 root_dir: str = 'content', sync: bool = True):
        self.resource_group = resource_group
        self.root_dir = root_dir
        self.sync = sync

    def __call__(self, resource_id: int):
        return FileStorage(
            resource_group=self.resource_group,
            root_dir=self.root_dir,
            sync=self.sync,
            resource_id=resource_id
        )



import sys
import os
import re
import pickle
import subprocess
from collections import OrderedDict
from pathlib import Path
import shutil

import numpy as np
from pandas.api import types as pd_types

from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

from .proc_wrapper import pd_decode, pd_encode, encode, decode


class BYOMHandler(BaseMLEngine):

    name = 'byom'

    def _get_model_proxy(self):
        con_args = self.engine_storage.get_connection_args()
        code = self.engine_storage.file_get(con_args['code'])
        modules_str = self.engine_storage.file_get(con_args['modules'])

        return ModelWrapper(
            code=code,
            modules_str=modules_str,
            engine_id=self.engine_storage.integration_id
        )

    def create(self, target, df=None, args=None, **kwargs):
        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            raise RuntimeError('BYOM is disabled on cloud')

        model_proxy = self._get_model_proxy()

        model_state = model_proxy.train(df, target, args)

        self.model_storage.file_set('model', model_state)

        # TODO return columns?

        def convert_type(field_type):
            if pd_types.is_integer_dtype(field_type):
                return 'integer'
            elif pd_types.is_numeric_dtype(field_type):
                return 'float'
            elif pd_types.is_datetime64_any_dtype(field_type):
                return 'datetime'
            else:
                return 'categorical'

        columns = {
            target: convert_type(np.object)
        }

        self.model_storage.columns_set(columns)

    def predict(self, df, args=None):
        pred_args = args.get('predict_params', {})

        model_proxy = self._get_model_proxy()

        model_state = self.model_storage.file_get('model')

        pred_df = model_proxy.predict(df, model_state, pred_args)

        return pred_df

    def create_engine(self, connection_args):
        # check code and requirements

        model_proxy = self._get_model_proxy()

        try:
            model_proxy.check()
        except Exception as e:
            # remove venv
            model_proxy.remove_venv()

            raise e


class ModelWrapper:
    def __init__(self, code, modules_str, engine_id):
        self.code = code
        modules = self.parse_requirements(modules_str)

        self.env_path = None
        self.prepare_env(modules, engine_id)

    def prepare_env(self, modules, engine_id):
        config = Config()

        try:
            import virtualenv

            base_path = config.get('byom', {}).get('venv_path')
            if base_path is None:
                # create in root path
                base_path = Path(config.paths['root']) / 'venvs'

            self.env_path = base_path / f'env_{engine_id}'

            self.python_path = self.env_path / 'bin' / 'python'

            if self.env_path.exists():
                # already exists. it means requirements are already installed
                return

            # create
            virtualenv.cli_run(['-p', sys.executable, str(self.env_path)])
            log.logger.info(f"Created new environment: {self.env_path}")

            if len(modules) > 0:
                self.install_modules(modules)

        except Exception as e:
            log.logger.info("Can't create virtual environment. venv module should be installed")

            self.python_path = Path(sys.executable)

            # try to install modules everytime
            self.install_modules(modules)

    def remove_venv(self):
        if self.env_path is not None and self.env_path.exists():
            shutil.rmtree(str(self.env_path))

    def parse_requirements(self, requirements):
        # get requirements from string
        # they should be located at the top of the file, before code

        pattern = '^[\w\\[\\]-]+[=!<>\s]*[\d\.]*[,=!<>\s]*[\d\.]*$'
        modules = []
        for line in requirements.split(b'\n'):
            line = line.decode().strip()
            if line:
                if re.match(pattern, line):
                    modules.append(line)
                else:
                    raise Exception(f'Wrong requirement: {line}')

        is_pandas = any([m.lower().startswith('pandas') for m in modules])
        if not is_pandas:
            modules.append('pandas >=1.1.5,<=1.3.3')

        # for dataframe serialization
        modules.append('pyarrow==11.0.0')
        return modules

    def install_modules(self, modules):
        # install in current environment using pip

        pip_cmd = self.python_path.parent / 'pip'
        for module in modules:
            p = subprocess.Popen([pip_cmd, 'install', module], stderr=subprocess.PIPE)
            p.wait()
            if p.returncode != 0:
                raise Exception(f'Problem with installing module {module}: {p.stderr.read()}')

    def _run_command(self, params):
        params_enc = encode(params)

        wrapper_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'proc_wrapper.py')
        p = subprocess.Popen(
            [str(self.python_path), wrapper_path],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        p.stdin.write(params_enc)
        p.stdin.close()
        ret_enc = p.stdout.read()

        p.wait()

        try:
            ret = decode(ret_enc)
        except (pickle.UnpicklingError, EOFError):
            raise RuntimeError(p.stderr.read())
        return ret

    def check(self):

        params = {
            'method': 'check',
            'code': self.code,
        }
        return self._run_command(params)

    def train(self, df, target, args):
        params = {
            'method': 'train',
            'df': pd_encode(df),
            'code': self.code,
            'to_predict': target,
            'args': args,
        }

        model_state = self._run_command(params)
        return model_state

    def predict(self, df, model_state, args):

        params = {
            'method': 'predict',
            'code': self.code,
            'df': pd_encode(df),
            'model_state': model_state,
            'args': args,
        }
        pred_df = self._run_command(params)
        return pd_decode(pred_df)


connection_args = OrderedDict(
    code={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model code'
    },
    modules={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model requirements'
    }
)


import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import CreatePredictor, CreateJob, RetrainPredictor, FinetunePredictor,\
    CreateMLEngine
from mindsdb_sql.parser.ast import Identifier, OrderBy, Insert, TableColumn, Constant

import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.api.mongo.utilities import logger
from mindsdb.integrations.libs.response import HandlerStatusResponse
from mindsdb.api.mongo.responders.find import find_to_ast
from mindsdb.api.mongo.responders.aggregate import aggregate_to_ast
from mindsdb.api.mongo.classes.query_sql import run_sql_command
from mindsdb.api.mongo.utilities.mongodb_parser import MongodbParser


class Responce(Responder):
    when = {'insert': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            res = self._result(query, request_env, mindsdb_env)
        except Exception as e:
            logger.error(e)
            res = {
                'n': 0,
                'writeErrors': [{
                    'index': 0,
                    'code': 0,
                    'errmsg': str(e)
                }],
                'ok': 1
            }
        return res

    def _insert_database(self, query, request_env, mindsdb_env):

        for doc in query['documents']:
            if '_id' in doc:
                del doc['_id']
            for field in ('name', 'engine', 'connection_args'):
                if field not in doc:
                    raise Exception(f"'{field}' must be specified")

            status = HandlerStatusResponse(success=False)
            try:
                handler = mindsdb_env['integration_controller'].create_tmp_handler(
                    handler_type=doc['engine'],
                    connection_data=doc['connection_args']
                )
                status = handler.check_connection()
            except Exception as e:
                status.error_message = str(e)

            if status.success is False:
                raise Exception(f"Can't connect to db: {status.error_message}")

            integration = mindsdb_env['integration_controller'].get(doc['name'])
            if integration is not None:
                raise Exception(f"Database '{doc['name']}' already exists.")

        for doc in query['documents']:
            mindsdb_env['integration_controller'].add(doc['name'], doc['engine'], doc['connection_args'])

    def _insert_model(self, query, request_env, mindsdb_env):
        predictors_columns = [
            'name',
            'status',
            'accuracy',
            'predict',
            'select_data_query',
            'training_options',
            'connection'
        ]

        if len(query['documents']) != 1:
            raise Exception("Must be inserted just one predictor at time")

        for doc in query['documents']:
            if '_id' in doc:
                del doc['_id']

            action = doc.pop('action', 'create').lower()

            bad_columns = [x for x in doc if x not in predictors_columns]
            if len(bad_columns) > 0:
                raise Exception(
                    f"Is no possible insert this columns to 'predictors' collection: {', '.join(bad_columns)}")

            if 'name' not in doc:
                raise Exception("Please, specify 'name' field")

            if 'predict' not in doc:
                predict = None
                if action == 'create':
                    raise Exception("Please, specify 'predict' field")
            else:
                predict = doc['predict']
                if not isinstance(predict, list):
                    predict = [Identifier(x.strip()) for x in predict.split(',')]

            order_by = None
            group_by = None
            ts_settings = {}

            kwargs = doc.get('training_options', {})

            if 'timeseries_settings' in kwargs:
                ts_settings = kwargs.pop('timeseries_settings')

                # mongo shell client sends int as float. need to convert it to int
                for key in ('window', 'horizon'):
                    val = ts_settings.get(key)
                    if val is not None:
                        ts_settings[key] = int(val)

                if 'order_by' in ts_settings:
                    order_by = ts_settings['order_by']
                    if not isinstance(order_by, list):
                        order_by = [order_by]

                    order_by = [
                        OrderBy(Identifier(x))
                        for x in order_by
                    ]
                if 'group_by' in ts_settings:
                    group_by = [
                        Identifier(x)
                        for x in ts_settings.get('group_by', [])
                    ]

            using = dict(kwargs)

            select_data_query = doc.get('select_data_query')
            integration_name = None
            if 'connection' in doc:
                integration_name = Identifier(doc['connection'])

            Class = CreatePredictor
            if action == 'retrain':
                Class = RetrainPredictor
            elif action == 'finetune':
                Class = FinetunePredictor

            create_predictor_ast = Class(
                name=Identifier(f"{request_env['database']}.{doc['name']}"),
                integration_name=integration_name,
                query_str=select_data_query,
                targets=predict,
                order_by=order_by,
                group_by=group_by,
                window=ts_settings.get('window'),
                horizon=ts_settings.get('horizon'),
                using=using,
            )

            run_sql_command(request_env, create_predictor_ast)

    def _insert_job(self, query, request_env, mindsdb_env):
        for doc in query['documents']:

            query_str = doc['query']
            # try parse as mongo
            parser = MongodbParser()
            try:
                mql = parser.from_string(query_str)

                method = mql.pipeline[0]['method']
                if method == 'find':
                    args = mql.pipeline[0]['args']
                    projection = None
                    if len(args) > 1:
                        projection = args[1]
                    query = {
                        'find': mql.collection,
                        'filter': args[0],
                        'projection': projection
                    }

                    for step in mql.pipeline[1:]:
                        if step['method'] == 'limit':
                            query['limit'] = step['args'][0]
                        if step['method'] == 'skip':
                            query['skip'] = step['args'][0]
                        if step['method'] == 'sort':
                            query['sort'] = step['args'][0]
                    # TODO implement group modifiers
                    ast_query = find_to_ast(query, request_env.get('database', 'mindsdb'))

                    # to string
                    query_str = ast_query.to_string()
                elif method == 'aggregate':
                    query = {
                        'aggregate': mql.collection,
                        'pipeline': mql.pipeline[0]['args'][0]
                    }
                    ast_query = aggregate_to_ast(query, request_env.get('database', 'mindsdb'))
                    query_str = ast_query.to_string()
            except Exception:
                # keep query
                pass

            repeat_str = doc.get('schedule_str').lower().lstrip('every ')

            ast_query = CreateJob(
                name=Identifier(doc['name']),
                start_str=doc.get('start_at'),
                end_str=doc.get('end_at'),
                query_str=query_str,
                repeat_str=repeat_str,
            )
            run_sql_command(request_env, ast_query)

    def _insert_ml_enine(self, query, request_env, mindsdb_env):
        for doc in query['documents']:

            ast_query = CreateMLEngine(
                name=Identifier(doc['name']),
                handler=doc['handler'],
                params=doc['params']
            )

            run_sql_command(request_env, ast_query)

    def _result(self, query, request_env, mindsdb_env):
        table = query['insert']

        if table == 'databases':
            self._insert_database(query, request_env, mindsdb_env)

        elif table in ['predictors', 'models']:
            self._insert_model(query, request_env, mindsdb_env)

        elif table == 'jobs':
            self._insert_job(query, request_env, mindsdb_env)

        elif table == 'ml_engines':
            self._insert_ml_enine(query, request_env, mindsdb_env)

        else:
            # regular insert

            df = pd.DataFrame(query['documents'])
            if '_id' in df.columns:
                df = df.drop('_id', axis=1)

            data = df.to_dict('split')
            values = []
            for row in data['data']:
                values.append([Constant(i) for i in row])
            ast_query = Insert(
                table=Identifier(table),
                columns=[TableColumn(c) for c in data['columns']],
                values=values
            )

            run_sql_command(request_env, ast_query)

        result = {
            "n": len(query['documents']),
            "ok": 1
        }

        return result


responder = Responce()



from io import BytesIO, StringIO
import os
import csv
import json
import codecs
import traceback
import tempfile
from urllib.parse import urlparse

       
import magic
import requests
import pandas as pd
from charset_normalizer import from_bytes

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import DropTables, Select

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


def clean_row(row):
    n_row = []
    for cell in row:
        if str(cell) in ['', ' ', '  ', 'NaN', 'nan', 'NA']:
            n_row.append(None)
        else:
            n_row.append(cell)

    return n_row


class FileHandler(DatabaseHandler):
    """
    Handler for files
    """
    name = 'files'

    def __init__(self, name=None, file_storage=None, connection_data={}, file_controller=None, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.fs_store = file_storage
        self.custom_parser = connection_data.get('custom_parser', None)
        self.clean_rows = connection_data.get('clean_rows', True)
        self.file_controller = file_controller

    def connect(self, **kwargs):
        return

    def disconnect(self, **kwargs):
        return

    def check_connection(self) -> StatusResponse:
        return StatusResponse(True)

    def query(self, query: ASTNode) -> Response:
        if type(query) == DropTables:
            for table_identifier in query.tables:
                if len(table_identifier.parts) == 2 and table_identifier.parts[0] != self.name:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table from database '{table_identifier.parts[0]}'"
                    )
                table_name = table_identifier.parts[-1]
                try:
                    self.file_controller.delete_file(table_name)
                except Exception as e:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table '{table_name}': {e}"
                    )
            return Response(RESPONSE_TYPE.OK)
        elif type(query) == Select:
            table_name = query.from_table.parts[-1]
            file_path = self.file_controller.get_file_path(table_name)
            df, _columns = self._handle_source(file_path, self.clean_rows, self.custom_parser)
            result_df = query_df(df, query)
            return Response(
                RESPONSE_TYPE.TABLE,
                data_frame=result_df
            )
        else:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message="Only 'select' and 'drop' queries allowed for files"
            )

    def native_query(self, query: str) -> Response:
        ast = self.parser(query, dialect='mindsdb')
        return self.query(ast)

    @staticmethod
    def _handle_source(file_path, clean_rows=True, custom_parser=None):
        # get file data io, format and dialect
        data, fmt, dialect = FileHandler._get_data_io(file_path)
        data.seek(0)  # make sure we are at 0 in file pointer

        if custom_parser:
            header, file_data = custom_parser(data, fmt)

        elif fmt == 'parquet':
            df = pd.read_parquet(data)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        elif fmt == 'csv':
            df = pd.read_csv(data, sep=dialect.delimiter)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        elif fmt in ['xlsx', 'xls']:
            data.seek(0)
            df = pd.read_excel(data)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        elif fmt == 'json':
            data.seek(0)
            json_doc = json.loads(data.read())
            df = pd.json_normalize(json_doc, max_level=0)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        else:
            raise ValueError('Could not load file into any format, supported formats are csv, json, xls, xlsx')

        if clean_rows:
            file_list_data = [clean_row(row) for row in file_data]
        else:
            file_list_data = file_data

        header = [x.strip() for x in header]
        col_map = dict((col, col) for col in header)
        return pd.DataFrame(file_list_data, columns=header), col_map

    @staticmethod
    def _get_data_io(file_path):
        """
        @TODO: Use python-magic to simplify the function and detect the file types as the xlsx example
        This gets a file either url or local file and defines what the format is as well as dialect
        :param file: file path or url
        :return: data_io, format, dialect
        """

        ############
        # get file as io object
        ############

        # file_path = self._get_file_path()

        data = BytesIO()

        try:
            with open(file_path, 'rb') as fp:
                data = BytesIO(fp.read())
        except Exception as e:
            error = 'Could not load file, possible exception : {exception}'.format(exception=e)
            print(error)
            raise ValueError(error)

        dialect = None

        ############
        # check for file type
        ############

        # Check first and last 4 bytes equal to PAR1.
        # Refer: https://parquet.apache.org/docs/file-format/
        parquet_sig = b'PAR1'
        data.seek(0, 0)
        start_meta = data.read(4)
        data.seek(-4, 2)
        end_meta = data.read()
        data.seek(0)
        if start_meta == parquet_sig and end_meta == parquet_sig:
            return data, 'parquet', None
        else:
            print("It's not parquet file. Checking for other formats")

        file_type = magic.from_file(file_path, mime=True)
        if file_type in ['application/vnd.openxmlformats-officedocument.spreadsheetml.sheet', 'application/vnd.ms-excel']:
            #for now we send xlsx for all excel types since we use pd.read_excel
            return data, 'xlsx', dialect
       
        # if not excel it can be a json file or a CSV, convert from binary to stringio

        byte_str = data.read()
        # Move it to StringIO
        try:
            # Handle Microsoft's BOM "special" UTF-8 encoding
            if byte_str.startswith(codecs.BOM_UTF8):
                data = StringIO(byte_str.decode('utf-8-sig'))
            else:
                file_encoding_meta = from_bytes(
                    byte_str[:32 * 1024],
                    steps=32,           # Number of steps/block to extract from my_byte_str
                    chunk_size=1024,    # Set block size of each extraction)
                    explain=False
                )
                best_meta = file_encoding_meta.best()
                errors = 'strict'
                if best_meta is not None:
                    encoding = file_encoding_meta.best().encoding
                else:
                    encoding = 'utf-8'
                    errors = 'replace'
                data = StringIO(byte_str.decode(encoding, errors))
        except Exception:
            print(traceback.format_exc())
            print('Could not load into string')

        # see if its JSON
        buffer = data.read(100)
        data.seek(0)
        text = buffer.strip()
        # analyze first n characters
        if len(text) > 0:
            text = text.strip()
            # it it looks like a json, then try to parse it
            if text.startswith('{') or text.startswith('['):
                try:
                    json.loads(data.read())
                    data.seek(0)
                    return data, 'json', dialect
                except Exception:
                    data.seek(0)
                    return data, None, dialect

        # lets try to figure out if its a csv
        try:
            dialect = FileHandler._get_csv_dialect(data)
            if dialect:
                return data, 'csv', dialect
            return data, None, dialect
        except Exception:
            data.seek(0)
            print('Could not detect format for this file')
            print(traceback.format_exc())
            # No file type identified
            return data, None, dialect

    @staticmethod
    def _get_file_path(path) -> str:
        try:
            is_url = urlparse(path).scheme in ('http', 'https')
        except Exception:
            is_url = False
        if is_url:
            path = FileHandler._fetch_url(path)
        return path

    @staticmethod
    def _get_csv_dialect(buffer) -> csv.Dialect:
        sample = buffer.readline()  # trying to get dialect from header
        buffer.seek(0)
        try:
            if isinstance(sample, bytes):
                sample = sample.decode()
            accepted_csv_delimiters = [',', '\t', ';']
            dialect = csv.Sniffer().sniff(sample, delimiters=accepted_csv_delimiters)
            dialect.doublequote = True  # assume that all csvs have " as string escape
        except csv.Error:
            dialect = None
        return dialect

    @staticmethod
    def _fetch_url(url: str) -> str:
        temp_dir = tempfile.mkdtemp(prefix='mindsdb_file_url_')
        try:
            r = requests.get(url, stream=True)
            if r.status_code == 200:
                with open(os.path.join(temp_dir, 'file'), 'wb') as f:
                    for chunk in r:
                        f.write(chunk)
            else:
                raise Exception(f'Responce status code is {r.status_code}')
        except Exception as e:
            print(f'Error during getting {url}')
            print(e)
            raise
        return os.path.join(temp_dir, 'file')

    def get_tables(self) -> Response:
        """
        List all files
        """
        files_meta = self.file_controller.get_files()
        data = [{
            'TABLE_NAME': x['name'],
            'TABLE_ROWS': x['row_count'],
            'TABLE_TYPE': 'BASE TABLE'
        } for x in files_meta]
        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(data)
        )

    def get_columns(self, table_name) -> Response:
        file_meta = self.file_controller.get_file_meta(table_name)
        result = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame([
                {
                    'Field': x['name'].strip() if isinstance(x, dict) else x.strip(),
                    'Type': 'str'
                } for x in file_meta['columns']
            ])
        )
        return result



import os
from setuptools import setup, find_packages


class Deps:
    pkgs = []
    pkgs_exclude = ["tests", "tests.*"]
    new_links = []
    extras = {}


about = {}
with open("mindsdb/__about__.py") as fp:
    exec(fp.read(), about)


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()


def define_deps():
    """Reads requirements.txt requirements-extra.txt files and preprocess it
    to be feed into setuptools.

    This is the only possible way (we found)
    how requirements.txt can be reused in setup.py
    using dependencies from private github repositories.

    Links must be appendend by `-{StringWithAtLeastOneNumber}`
    or something like that, so e.g. `-9231` works as well as
    `1.1.0`. This is ignored by the setuptools, but has to be there.

    Warnings:
        to make pip respect the links, you have to use
        `--process-dependency-links` switch. So e.g.:
        `pip install --process-dependency-links {git-url}`

    Returns:
         list of packages, extras and dependency links.
    """
    with open(os.path.normpath('requirements/requirements.txt')) as req_file:
        defaults = [req.strip() for req in req_file.read().splitlines()]

    links = []
    requirements = []
    for r in defaults:
        if 'git+https' in r:
            pkg = r.split('#')[-1]
            links.append(r + '-9876543210')
            requirements.append(pkg.replace('egg=', ''))
        else:
            requirements.append(r.strip())

    extra_requirements = {}
    full_requirements = []
    for fn in os.listdir(os.path.normpath('./requirements')):
        if fn.startswith('requirements-') and fn.endswith('.txt'):
            extra_name = fn.replace('requirements-', '').replace('.txt', '')
            with open(os.path.normpath(f"./requirements/{fn}")) as fp:
                extra = [req.strip() for req in fp.read().splitlines()]
            extra_requirements[extra_name] = extra
            full_requirements += extra

    extra_requirements['all_extras'] = list(set(full_requirements))

    full_handlers_requirements = []
    handlers_dir_path = os.path.normpath('./mindsdb/integrations/handlers')
    for fn in os.listdir(handlers_dir_path):
        if os.path.isdir(os.path.join(handlers_dir_path, fn)) and fn.endswith("_handler"):
            req_file_path = os.path.join(handlers_dir_path, fn, 'requirements.txt')
            if os.path.exists(req_file_path):
                with open(req_file_path) as fp:
                    extra = [req.strip() for req in fp.read().splitlines()]
                extra_name = fn.replace("_handler", "")
                extra_requirements[extra_name] = extra
                full_handlers_requirements += extra

    extra_requirements['all_handlers_extras'] = list(set(full_handlers_requirements))

    Deps.pkgs = requirements
    Deps.extras = extra_requirements
    Deps.new_links = links

    return Deps


deps = define_deps()

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=deps.pkgs_exclude),
    install_requires=deps.pkgs,
    dependency_links=deps.new_links,
    extras_require=deps.extras,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8"
)



from typing import Union, BinaryIO

from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message_identifiers import \
    PostgresBackendMessageIdentifier, PostgresFrontendMessageIdentifier
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_packets import PostgresPacketReader, \
    PostgresPacketBuilder


class PostgresMessage:
    identifier: Union[PostgresBackendMessageIdentifier, PostgresFrontendMessageIdentifier]
    backend_capable: bool
    frontend_capable: bool

    def __init__(self):
        pass

    def send(self, write_file: BinaryIO):
        # with open("test_write.txt", "a") as f:
        #    f.write(str(type(self)))
        return self.send_internal(write_file=write_file)

    def send_internal(self, write_file: BinaryIO):
        raise NotImplementedError("Must implement send_internal in sub-class")

    def read(self, packet_reader: PostgresPacketReader):
        pass

    def get_packet_builder(self) -> PostgresPacketBuilder:
        ppb = PostgresPacketBuilder()
        ppb.set_identifier(self.identifier)
        return ppb



from ..clickhouse_handler.clickhouse_handler import ClickHouseHandler

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from collections import OrderedDict
from typing import Optional


class DatabendHandler(ClickHouseHandler):
    """
    This handler handles connection and execution of the Databend statements.
    """
    name = 'databend'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        super().__init__(name, connection_data, **kwargs)


connection_args = OrderedDict(
    protocol={
        'type': ARG_TYPE.STR,
        'protocol': 'The protocol to query Databend. Supported: native, http, https. Default: native'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Databend warehouse.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Databend warehouse.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Databend warehouse.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Databend warehouse. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the ClickHouse server.'
    }
)

connection_args_example = OrderedDict(
    protocol='https',
    host='some-url.aws-us-east-2.default.databend.com',
    port=443,
    user='root',
    password='password',
    database='test_db'
)


import pandas as pd
from typing import Text, List, Dict

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities.select_query_utilities import SELECTQueryParser, SELECTQueryExecutor


class CustomersTable(APITable):
    """The Stripe Customers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Customer data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Customers matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'customers',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            customers_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        customers_df = select_statement_executor.execute_query()

        return customers_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_customers(limit=1)).columns.tolist()

    def get_customers(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        customers = stripe.Customer.list(**kwargs)
        return [customer.to_dict() for customer in customers]


class ProductsTable(APITable):
    """The Stripe Products Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Product data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Products matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'products',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            products_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        products_df = select_statement_executor.execute_query()

        return products_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_products(limit=1)).columns.tolist()

    def get_products(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        products = stripe.Product.list(**kwargs)
        return [product.to_dict() for product in products]


class PaymentIntentsTable(APITable):
    """The Stripe Payment Intents Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Payment Intents data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Payment Intents matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payment_intents',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        payment_intents_df = pd.json_normalize(self.get_payment_intents(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            payment_intents_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        payment_intents_df = select_statement_executor.execute_query()

        return payment_intents_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payment_intents(limit=1)).columns.tolist()

    def get_payment_intents(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        payment_intents = stripe.PaymentIntent.list(**kwargs)
        return [payment_intent.to_dict() for payment_intent in payment_intents]



import os
import json
from pathlib import Path
import shutil

from mindsdb.interfaces.storage import db
from mindsdb.integrations.handlers.file_handler import Handler as FileHandler
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities.context import context as ctx


class FileController():
    def __init__(self):
        self.config = Config()
        self.fs_store = FsStore()
        self.dir = os.path.join(self.config.paths['content'], 'file')

    def get_files_names(self):
        """ return list of files names
        """
        return [x[0] for x in db.session.query(db.File.name).filter_by(company_id=ctx.company_id)]

    def get_file_meta(self, name):
        file_record = db.session.query(db.File).filter_by(
            company_id=ctx.company_id,
            name=name
        ).first()
        if file_record is None:
            return None
        columns = file_record.columns
        if isinstance(columns, str):
            columns = json.loads(columns)
        return {
            'name': file_record.name,
            'columns': columns,
            'row_count': file_record.row_count
        }

    def get_files(self):
        """ Get list of files

            Returns:
                list[dict]: files metadata
        """
        file_records = db.session.query(db.File).filter_by(company_id=ctx.company_id).all()
        files_metadata = [{
            'name': record.name,
            'row_count': record.row_count,
            'columns': record.columns,
        } for record in file_records]
        return files_metadata

    def save_file(self, name, file_path, file_name=None):
        """ Save the file to our store

            Args:
                name (str): with that name file will be available in sql api
                file_name (str): file name
                file_path (str): path to the file

            Returns:
                int: id of 'file' record in db
        """
        files_metadata = self.get_files()
        if name in [x['name'] for x in files_metadata]:
            raise Exception(f'File already exists: {name}')

        if file_name is None:
            file_name = Path(file_path).name

        file_dir = None
        try:
            df, _col_map = FileHandler._handle_source(file_path)

            ds_meta = {
                'row_count': len(df),
                'column_names': list(df.columns)
            }

            file_record = db.File(
                name=name,
                company_id=ctx.company_id,
                source_file_path=file_name,
                file_path='',
                row_count=ds_meta['row_count'],
                columns=ds_meta['column_names']
            )
            db.session.add(file_record)
            db.session.commit()
            store_file_path = f'file_{ctx.company_id}_{file_record.id}'
            file_record.file_path = store_file_path
            db.session.commit()

            file_dir = Path(self.dir).joinpath(store_file_path)
            file_dir.mkdir(parents=True, exist_ok=True)
            source = file_dir.joinpath(file_name)
            # NOTE may be delay between db record exists and file is really in folder
            shutil.move(file_path, str(source))

            self.fs_store.put(store_file_path, base_dir=self.dir)
        except Exception as e:
            log.logger.error(e)
            raise
        finally:
            if file_dir is not None:
                shutil.rmtree(file_dir)

        return file_record.id

    def delete_file(self, name):
        file_record = db.session.query(db.File).filter_by(company_id=ctx.company_id, name=name).first()
        if file_record is None:
            return None
        file_id = file_record.id
        db.session.delete(file_record)
        db.session.commit()
        self.fs_store.delete(f'file_{ctx.company_id}_{file_id}')
        return True

    def get_file_path(self, name):
        file_record = db.session.query(db.File).filter_by(company_id=ctx.company_id, name=name).first()
        if file_record is None:
            raise Exception(f"File '{name}' does not exists")
        file_dir = f'file_{ctx.company_id}_{file_record.id}'
        self.fs_store.get(file_dir, base_dir=self.dir)
        return str(Path(self.dir).joinpath(file_dir).joinpath(Path(file_record.source_file_path).name))



import re

import mindsdb.interfaces.storage.db as db

from .fs import RESOURCE_GROUP, FileStorageFactory
from .json import get_json_storage


class ModelStorage:
    """
    This class deals with all model-related storage requirements, from setting status to storing artifacts.
    """
    def __init__(self, predictor_id):
        storageFactory = FileStorageFactory(
            resource_group=RESOURCE_GROUP.PREDICTOR,
            sync=True
        )
        self.fileStorage = storageFactory(predictor_id)
        self.predictor_id = predictor_id

    # -- fields --

    def get_info(self):
        rec = db.Predictor.query.get(self.predictor_id)
        return dict(status=rec.status,
                    to_predict=rec.to_predict,
                    data=rec.data)

    def status_set(self, status, status_info=None):
        rec = db.Predictor.query.get(self.predictor_id)
        rec.status = status
        if status_info is not None:
            rec.data = status_info
        db.session.commit()

    def training_state_set(self, current_state_num=None, total_states=None, state_name=None):
        rec = db.Predictor.query.get(self.predictor_id)
        if current_state_num is not None:
            rec.training_phase_current = current_state_num
        if total_states is not None:
            rec.training_phase_total = total_states
        if state_name is not None:
            rec.training_phase_name = state_name
        db.session.commit()

    def training_state_get(self):
        rec = db.Predictor.query.get(self.predictor_id)
        return [rec.training_phase_current, rec.training_phase_total, rec.training_phase_name]

    def columns_get(self):
        rec = db.Predictor.query.get(self.predictor_id)
        return rec.dtype_dict

    def columns_set(self, columns):
        # columns: {name: dtype}

        rec = db.Predictor.query.get(self.predictor_id)
        rec.dtype_dict = columns
        db.session.commit()

    # files

    def file_get(self, name):
        return self.fileStorage.file_get(name)

    def file_set(self, name, content):
        self.fileStorage.file_set(name, content)

    def file_list(self):
        ...

    def file_del(self, name):
        ...

    # jsons

    def json_set(self, name, data):
        json_storage = get_json_storage(
            resource_id=self.predictor_id,
            resource_group=RESOURCE_GROUP.PREDICTOR
        )
        return json_storage.set(name, data)

    def json_get(self, name):
        json_storage = get_json_storage(
            resource_id=self.predictor_id,
            resource_group=RESOURCE_GROUP.PREDICTOR
        )
        return json_storage.get(name)

    def json_list(self):
        ...

    def json_del(self, name):
        ...

    def delete(self):
        self.fileStorage.delete()
        # TODO delete json


class HandlerStorage:
    """
    This class deals with all handler-related storage requirements, from storing metadata to synchronizing folders
    across instances.
    """
    def __init__(self, integration_id):
        storageFactory = FileStorageFactory(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            sync=False
        )
        self.fileStorage = storageFactory(integration_id)
        self.integration_id = integration_id

    def get_connection_args(self):
        rec = db.Integration.query.get(self.integration_id)
        return rec.data

    # files

    def file_get(self, name):
        self.fileStorage.pull_path(name)
        return self.fileStorage.file_get(name)

    def file_set(self, name, content):
        self.fileStorage.file_set(name, content)
        self.fileStorage.push_path(name)

    def file_list(self):
        ...

    def file_del(self, name):
        ...

    # folder

    def folder_get(self, name, update=True):
        # pull folder and return path
        name = name.lower().replace(' ', '_')
        name = re.sub(r'([^a-z^A-Z^_\d]+)', '_', name)

        self.fileStorage.pull_path(name, update=update)
        return str(self.fileStorage.get_path(name))

    def folder_sync(self, name):
        # sync abs path
        name = name.lower().replace(' ', '_')
        name = re.sub(r'([^a-z^A-Z^_\d]+)', '_', name)

        self.fileStorage.push_path(name)

    # jsons

    def json_set(self, name, content):
        ...

    def json_get(self, name):
        ...

    def json_list(self):
        ...

    def json_del(self, name):
        ...



import re
import os
from typing import Optional, Dict

import numpy as np
import pandas as pd

from langchain.schema import SystemMessage
from langchain.agents import AgentType
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI  # GPT-4 fails to follow the output langchain requires, avoid using for now
from langchain.agents import initialize_agent, load_tools, Tool, create_sql_agent
from langchain.prompts import PromptTemplate
from langchain.utilities import GoogleSerperAPIWrapper
from langchain.agents.agent_toolkits import SQLDatabaseToolkit
from langchain.chains.conversation.memory import ConversationSummaryBufferMemory

from mindsdb.integrations.handlers.openai_handler.openai_handler import OpenAIHandler, CHAT_MODELS
from mindsdb.integrations.handlers.langchain_handler.mindsdb_database_agent import MindsDBSQL
from mindsdb_sql import parse_sql, Insert


_DEFAULT_MODEL = 'gpt-3.5-turbo'  # TODO: enable other LLM backends (AI21, Anthropic, etc.)
_DEFAULT_MAX_TOKENS = 2048  # requires more than vanilla OpenAI due to ongoing summarization and 3rd party input
_DEFAULT_AGENT_MODEL = 'zero-shot-react-description'
_DEFAULT_AGENT_TOOLS = ['python_repl', 'wikipedia']  # these require no additional arguments


class LangChainHandler(OpenAIHandler):
    """
    This is a MindsDB integration for the LangChain library, which provides a unified interface for interacting with
    various large language models (LLMs).

    Currently, this integration supports exposing OpenAI's LLMs with normal text completion support. They are then
    wrapped in a zero shot react description agent that offers a few third party tools out of the box, with support
    for additional ones if an API key is provided. Ongoing memory is also provided.

    Full tool support list:
        - wikipedia
        - python_repl
        - serper.dev search

    This integration inherits from the OpenAI engine, so it shares a lot of the requirements, features (e.g. prompt
    templating) and limitations.
    """
    name = 'langchain'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.stops = []
        self.default_model = _DEFAULT_MODEL
        self.default_max_tokens = _DEFAULT_MAX_TOKENS
        self.default_agent_model = _DEFAULT_AGENT_MODEL
        self.default_agent_tools = _DEFAULT_AGENT_TOOLS
        self.write_privileges = False  # if True, this agent is able to write into other active mindsdb integrations

    def _get_serper_api_key(self, args, strict=True):
        if 'serper_api_key' in args:
            return args['serper_api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'serper_api_key' in connection_args:
            return connection_args['serper_api_key']
        # 3
        api_key = os.getenv('SERPER_API_KEY')  # e.g. "OPENAI_API_KEY"
        if api_key is not None:
            return api_key

        if strict:
            raise Exception(f'Missing API key serper_api_key. Either re-create this ML_ENGINE specifying the `serper_api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  # noqa

    def create(self, target, args=None, **kwargs):
        self.write_privileges = args.get('using', {}).get('writer', self.write_privileges)
        self.default_agent_tools = args.get('tools', self.default_agent_tools)
        super().create(target, args, **kwargs)

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("LangChain engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        if args.get('mode') != 'conversational':
            if len(set(args.keys()) & {'prompt_template'}) == 0:
                raise Exception('Please provide a `prompt_template` for this engine.')

    def predict(self, df, args=None):
        """
        Dispatch is performed depending on the underlying model type. Currently, only the default text completion
        is supported.
        """
        executor = args['executor']  # used as tool in custom tool for the agent to have mindsdb-wide access
        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        args['executor'] = executor

        df = df.reset_index(drop=True)

        if args.get('mode') != 'conversational':
            if 'prompt_template' not in args and 'prompt_template' not in pred_args:
                raise Exception(f"This model expects a prompt template, please provide one.")

        if 'stops' in pred_args:
            self.stops = pred_args['stops']

        # TODO: offload creation to the `create` method instead for faster inference?

        modal_dispatch = {
            'default': 'default_completion',
            'sql_agent': 'sql_agent_completion',
        }

        agent_creation_method = modal_dispatch.get(args.get('modal_dispatch', 'default'), 'default_completion')

        if args.get('mode') == 'conversational':
            agent_creation_method = 'conversational_completion'

        # input dataframe can be modified
        agent, df = getattr(self, agent_creation_method)(df, args, pred_args)
        return self.run_agent(df, agent, args, pred_args)

    def conversational_completion(self, df, args=None, pred_args=None):
        pred_args = pred_args if pred_args else {}

        # api argument validation
        model_name = args.get('model_name', self.default_model)

        model_kwargs = {
            'model_name': model_name,
            'temperature': min(1.0, max(0.0, args.get('temperature', 0.0))),
            'max_tokens': pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens)),
            'top_p': pred_args.get('top_p', None),
            'frequency_penalty': pred_args.get('frequency_penalty', None),
            'presence_penalty': pred_args.get('presence_penalty', None),
            'n': pred_args.get('n', None),
            'best_of': pred_args.get('best_of', None),
            'request_timeout': pred_args.get('request_timeout', None),
            'logit_bias': pred_args.get('logit_bias', None),
            'openai_api_key': self._get_openai_api_key(args, strict=True),
            'serper_api_key': self._get_serper_api_key(args, strict=False),
        }
        model_kwargs = {k: v for k, v in model_kwargs.items() if v is not None}  # filter out None values

        # langchain tool setup
        tools = self._setup_tools(model_kwargs, pred_args, args['executor'])

        llm = ChatOpenAI(**model_kwargs)

        memory = ConversationSummaryBufferMemory(llm=llm,
                                                 max_token_limit=model_kwargs.get('max_tokens', None),
                                                 memory_key="chat_history")

        # fill memory

        # system prompt
        prompt = args['prompt']
        if 'prompt' in pred_args:
            prompt = pred_args['prompt']
        if 'context' in pred_args:
            prompt += '\n\n' + 'Useful information:\n' + pred_args['context'] + '\n'
        memory.chat_memory.messages.insert(0, SystemMessage(content=prompt))

        # user - assistant conversation. get all except the last message
        for row in df[:-1].to_dict('records'):
            question = row[args['user_column']]
            answer = row[args['assistant_column']]

            if question:
                memory.chat_memory.add_user_message(question)
            if answer:
                memory.chat_memory.add_ai_message(answer)

        # use last message as prompt, remove other questions
        df.iloc[:-1, df.columns.get_loc('question')] = ''

        agent_name = AgentType.CONVERSATIONAL_REACT_DESCRIPTION
        agent = initialize_agent(
            tools,
            llm,
            memory=memory,
            agent=agent_name,
            max_iterations=pred_args.get('max_iterations', 3),
            verbose=pred_args.get('verbose', args.get('verbose', False)),
            handle_parsing_errors=True,
        )

        # setup model description
        description = {
            'allowed_tools': [agent.agent.allowed_tools],  # packed as list to avoid additional rows
            'agent_type': agent_name,
            'max_iterations': agent.max_iterations,
            'memory_type': memory.__class__.__name__,
        }

        description = {**description, **model_kwargs}
        description.pop('openai_api_key', None)
        self.model_storage.json_set('description', description)

        return agent, df

    def default_completion(self, df, args=None, pred_args=None):
        """
        Mostly follows the logic of the OpenAI handler, but with a few additions:
            - setup the langchain toolkit
            - setup the langchain agent (memory included)
            - setup information to be published when describing the model

        Ref link from the LangChain documentation on how to accomplish the first two items:
            - python.langchain.com/en/latest/modules/agents/agents/custom_agent.html
        """
        pred_args = pred_args if pred_args else {}

        # api argument validation
        model_name = pred_args.get('model_name', args.get('model_name', self.default_model))
        agent_name = pred_args.get('agent_name', args.get('agent_name', self.default_agent_model))

        model_kwargs = {
            'model_name': model_name,
            'temperature': min(1.0, max(0.0, pred_args.get('temperature', args.get('temperature', 0.0)))),
            'max_tokens': pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens)),
            'top_p': pred_args.get('top_p', None),
            'frequency_penalty': pred_args.get('frequency_penalty', None),
            'presence_penalty': pred_args.get('presence_penalty', None),
            'n': pred_args.get('n', None),
            'best_of': pred_args.get('best_of', None),
            'request_timeout': pred_args.get('request_timeout', None),
            'logit_bias': pred_args.get('logit_bias', None),
            'openai_api_key': self._get_openai_api_key(args, strict=True),
            'serper_api_key': self._get_serper_api_key(args, strict=False),
        }
        model_kwargs = {k: v for k, v in model_kwargs.items() if v is not None}  # filter out None values

        # langchain tool setup
        pred_args['tools'] = args.get('tools') if 'tools' not in pred_args else pred_args.get('tools', [])
        tools = self._setup_tools(model_kwargs, pred_args, args['executor'])

        # langchain agent setup
        if model_kwargs['model_name'] in CHAT_MODELS:
            llm = ChatOpenAI(**model_kwargs)
        else:
            llm = OpenAI(**model_kwargs)
        memory = ConversationSummaryBufferMemory(llm=llm, max_token_limit=model_kwargs.get('max_tokens', None))
        agent = initialize_agent(
            tools,
            llm,
            memory=memory,
            agent=agent_name,
            max_iterations=pred_args.get('max_iterations', 3),
            verbose=pred_args.get('verbose', args.get('verbose', False)),
            handle_parsing_errors=True,
        )

        # setup model description
        description = {
            'allowed_tools': [agent.agent.allowed_tools],   # packed as list to avoid additional rows
            'agent_type': agent_name,
            'max_iterations': agent.max_iterations,
            'memory_type': memory.__class__.__name__,
        }
        description = {**description, **model_kwargs}
        description.pop('openai_api_key', None)
        self.model_storage.json_set('description', description)

        return agent, df

    def run_agent(self, df, agent, args, pred_args):
        # TODO abstract prompt templating into a common utility method, this is also used in vanilla OpenAI
        if args.get('prompt_template', False):
            base_template = args['prompt_template']  # override with predict-time template if available
        elif 'prompt_template' in pred_args:
            base_template = pred_args['prompt_template']
        else:
            base_template = '{{question}}'

        input_variables = []
        matches = list(re.finditer("{{(.*?)}}", base_template))

        for m in matches:
            input_variables.append(m[0].replace('{', '').replace('}', ''))

        empty_prompt_ids = np.where(df[input_variables].isna().all(axis=1).values)[0]

        base_template = base_template.replace('{{', '{').replace('}}', '}')
        prompts = []

        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                prompt = PromptTemplate(input_variables=input_variables, template=base_template)
                kwargs = {}
                for col in input_variables:
                    kwargs[col] = row[col] if row[col] is not None else ''  # add empty quote if data is missing
                prompts.append(prompt.format(**kwargs))

        def _completion(agent, prompts):
            # TODO: ensure that agent completion plus prompt match the maximum allowed by the user
            # TODO: use async API if possible for parallelized completion
            completions = []
            for prompt in prompts:
                if not prompt:
                    # skip empty values
                    completions.append('')
                    continue
                try:
                    completions.append(agent.run(prompt))
                except Exception as e:
                    completions.append(f'agent failed with error:\n{str(e)[:50]}...')
            return [c for c in completions]

        completion = _completion(agent, prompts)

        # add null completion for empty prompts
        for i in sorted(empty_prompt_ids):
            completion.insert(i, None)

        pred_df = pd.DataFrame(completion, columns=[args['target']])

        return pred_df

    def _setup_tools(self, model_kwargs, pred_args, executor):
        def _mdb_exec_call(query: str) -> str:
            """ We define it like this to pass the executor through the closure, as custom classes don't allow custom field assignment. """  # noqa
            try:
                ast_query = parse_sql(query.strip('`'), dialect='mindsdb')
                ret = executor.execute_command(ast_query)

                data = ret.data  # list of lists
                data = '\n'.join([  # rows
                    '\t'.join(      # columns
                        str(row) if isinstance(row, str) else [str(value) for value in row]
                    ) for row in data
                ])
            except Exception as e:
                data = f"mindsdb tool failed with error:\n{str(e)}"   # let the agent know
            return data

        def _mdb_exec_metadata_call(query: str) -> str:
            try:
                parts = query.replace('`', '').split('.')
                assert 1 <= len(parts) <= 2, 'query must be in the format: `integration` or `integration.table`'

                integration = parts[0]
                integrations = executor.session.integration_controller
                handler = integrations.get_handler(integration)

                if len(parts) == 1:
                    df = handler.get_tables().data_frame
                    data = f'The integration `{integration}` has {df.shape[0]} tables: {", ".join(list(df["TABLE_NAME"].values))}'  # noqa

                if len(parts) == 2:
                    df = handler.get_tables().data_frame
                    table_name = parts[-1]
                    try:
                        table_name_col = 'TABLE_NAME' if 'TABLE_NAME' in df.columns else 'table_name'
                        mdata = df[df[table_name_col] == table_name].iloc[0].to_list()
                        if len(mdata) == 3:
                            _, nrows, table_type = mdata
                            data = f'Metadata for table {table_name}:\n\tRow count: {nrows}\n\tType: {table_type}\n'
                        elif len(mdata) == 2:
                            nrows = mdata
                            data = f'Metadata for table {table_name}:\n\tRow count: {nrows}\n'
                        else:
                            data = f'Metadata for table {table_name}:\n'
                        fields = handler.get_columns(table_name).data_frame['Field'].to_list()
                        types = handler.get_columns(table_name).data_frame['Type'].to_list()
                        data += f'List of columns and types:\n'
                        data += '\n'.join([f'\tColumn: `{field}`\tType: `{typ}`' for field, typ in zip(fields, types)])
                    except:
                        data = f'Table {table_name} not found.'
            except Exception as e:
                data = f"mindsdb tool failed with error:\n{str(e)}"  # let the agent know
            return data

        def _mdb_write_call(query: str) -> str:
            try:
                query = query.strip('`')
                ast_query = parse_sql(query.strip('`'), dialect='mindsdb')
                if isinstance(ast_query, Insert):
                    _ = executor.execute_command(ast_query)
                    return "mindsdb write tool executed successfully"
            except Exception as e:
                return f"mindsdb write tool failed with error:\n{str(e)}"

        mdb_tool = Tool(
                name="MindsDB",
                func=_mdb_exec_call,
                description="useful to read from databases or tables connected to the mindsdb machine learning package. the action must be a valid simple SQL query, always ending with a semicolon. For example, you can do `show databases;` to list the available data sources, and `show tables;` to list the available tables within each data source."  # noqa
            )

        mdb_meta_tool = Tool(
            name="MDB-Metadata",
            func=_mdb_exec_metadata_call,
            description="useful to get column names from a mindsdb table or metadata from a mindsdb data source. the command should be either 1) a data source name, to list all available tables that it exposes, or 2) a string with the format `data_source_name.table_name` (for example, `files.my_table`), to get the table name, table type, column names, data types per column, and amount of rows of the specified table."  # noqa
        )

        mdb_write_tool = Tool(
            name="MDB-Write",
            func=_mdb_write_call,
            description="useful to write into data sources connected to mindsdb. command must be a valid SQL query with syntax: `INSERT INTO data_source_name.table_name (column_name_1, column_name_2, [...]) VALUES (column_1_value_row_1, column_2_value_row_1, [...]), (column_1_value_row_2, column_2_value_row_2, [...]), [...];`. note the command always ends with a semicolon. order of column names and values for each row must be a perfect match. If write fails, try casting value with a function, passing the value without quotes, or truncating string as needed.`."  # noqa
        )

        toolkit = pred_args['tools'] if pred_args['tools'] is not None else self.default_agent_tools

        standard_tools = []
        custom_tools = []
        # possible to pass standart tool name or custom function
        for tool in toolkit:
            if isinstance(tool, str):
                standard_tools.append(tool)
            else:
                custom_tools.append(tool)

        tools = load_tools(standard_tools)
        if model_kwargs.get('serper_api_key', False):
            search = GoogleSerperAPIWrapper(serper_api_key=model_kwargs.pop('serper_api_key'))
            tools.append(Tool(
                name="Intermediate Answer (serper.dev)",
                func=search.run,
                description="useful for when you need to search the internet (note: in general, use this as a last resort)"  # noqa
            ))

        # add connection to mindsdb
        tools.append(mdb_tool)
        tools.append(mdb_meta_tool)

        if self.write_privileges:
            tools.append(mdb_write_tool)

        for tool in custom_tools:
            tools.append(Tool(
                name=tool['name'],
                func=tool['func'],
                description=tool['description'],
            ))

        return tools

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        info = self.model_storage.json_get('description')

        if attribute == 'info':
            if info is None:
                # we do this due to the huge amount of params that can be changed
                #  at prediction time to customize behavior.
                # for them, we report the last observed value
                raise Exception('This model needs to be used before it can be described.')

            description = pd.DataFrame(info)
            return description
        else:
            tables = ['info']
            return pd.DataFrame(tables, columns=['tables'])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        raise NotImplementedError('Fine-tuning is not supported for LangChain models')

    def sql_agent_completion(self, df, args=None, pred_args=None):
        """This completion will be used to answer based on information passed by any MindsDB DB or API engine."""
        db = MindsDBSQL(engine=args['executor'], metadata=args['executor'].session.integration_controller)
        toolkit = SQLDatabaseToolkit(db=db)
        model_name = args.get('model_name', self.default_model)
        llm = OpenAI(temperature=0) if model_name not in CHAT_MODELS else ChatOpenAI(temperature=0)
        agent = create_sql_agent(
            llm=llm,
            toolkit=toolkit,
            verbose=True
        )
        return agent, df



import sys
import json
import copy
from typing import Optional, Dict
from datetime import datetime
from functools import lru_cache

import pandas as pd
from type_infer.dtype import dtype
import lightwood
import numpy as np

import mindsdb.interfaces.storage.db as db

from mindsdb.utilities.functions import cast_row_types
# from mindsdb.utilities.hooks import after_predict as after_predict_hook
from mindsdb.interfaces.model.functions import get_model_record
from mindsdb.interfaces.storage.json import get_json_storage
from mindsdb.integrations.libs.base import BaseMLEngine
import mindsdb.utilities.profiler as profiler

from .functions import run_learn, run_finetune

IS_PY36 = sys.version_info[1] <= 6


class NumpyJSONEncoder(json.JSONEncoder):
    """
    Use this encoder to avoid
    "TypeError: Object of type float32 is not JSON serializable"

    Example:
    x = np.float32(5)
    json.dumps(x, cls=NumpyJSONEncoder)
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float, np.float32, np.float64)):
            return float(obj)
        else:
            return super().default(obj)


class LightwoodHandler(BaseMLEngine):
    name = 'lightwood'

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'df' not in kwargs:
            return
        df = kwargs['df']
        columns = [x.lower() for x in df.columns]
        if target.lower() not in columns:
            raise Exception(f"There is no column '{target}' in dataframe")

        if 'timeseries_settings' in args and args['timeseries_settings'].get('is_timeseries') is True:
            tss = args['timeseries_settings']
            if 'order_by' in tss and tss['order_by'].lower() not in columns:
                raise Exception(f"There is no column '{tss['order_by']}' in dataframe")
            if isinstance(tss.get('group_by'), list):
                for column in tss['group_by']:
                    if column.lower() not in columns:
                        raise Exception(f"There is no column '{column}' in dataframe")

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        args['target'] = target
        run_learn(
            df,
            args,   # Problem definition and JsonAI override
            self.model_storage
        )

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        run_finetune(
            df,
            args,
            self.model_storage
        )

    @staticmethod
    @lru_cache(maxsize=5)
    def get_predictor(predictor_path, predictor_code):
        predictor = lightwood.predictor_from_state(predictor_path, predictor_code)
        return predictor

    @profiler.profile('LightwoodHandler.predict')
    def predict(self, df, args=None):
        pred_format = args['pred_format']
        predictor_code = args['code']
        learn_args = args['learn_args']
        pred_args = args.get('predict_params', {})
        self.model_storage.fileStorage.pull()

        with profiler.Context('load model'):
            predictor_path = self.model_storage.fileStorage.folder_path / self.model_storage.fileStorage.folder_name
            predictor = LightwoodHandler.get_predictor(predictor_path, predictor_code)

        dtype_dict = predictor.dtype_dict

        with profiler.Context('predict'):
            predictions = predictor.predict(df, args=pred_args)

        predictions = predictions.to_dict(orient='records')

        # TODO!!!
        # after_predict_hook(
        #     company_id=self.company_id,
        #     predictor_id=predictor_record.id,
        #     rows_in_count=df.shape[0],
        #     columns_in_count=df.shape[1],
        #     rows_out_count=len(predictions)
        # )

        # region format result
        target = args['target']
        explain_arr = []
        pred_dicts = []
        for i, row in enumerate(predictions):
            values = {
                'predicted_value': row['prediction'],
                'confidence': row.get('confidence', None),
                'anomaly': row.get('anomaly', None),
                'truth': row.get('truth', None)
            }

            if predictor.supports_proba:
                for cls in predictor.statistical_analysis.train_observed_classes:
                    if row.get(f'__mdb_proba_{cls}', False):
                        values[f'probability_class_{cls}'] = round(row[f'__mdb_proba_{cls}'], 4)

            for block in predictor.analysis_blocks:
                if type(block).__name__ == 'ShapleyValues':
                    cols = block.columns
                    values['shap_base_response'] = round(row['shap_base_response'], 4)
                    values['shap_final_response'] = round(row['shap_final_response'], 4)
                    for col in cols:
                        values[f'shap_contribution_{col}'] = round(row[f'shap_contribution_{col}'], 4)

            if 'lower' in row:
                values['confidence_lower_bound'] = row.get('lower', None)
                values['confidence_upper_bound'] = row.get('upper', None)

            obj = {target: values}
            explain_arr.append(obj)

            td = {'predicted_value': row['prediction']}
            for col in df.columns:
                if col in row:
                    td[col] = row[col]
                elif f'order_{col}' in row:
                    td[col] = row[f'order_{col}']
                elif f'group_{col}' in row:
                    td[col] = row[f'group_{col}']
                else:
                    orginal_index = row.get('original_index')
                    if orginal_index is None:
                        orginal_index = i
                    td[col] = df.iloc[orginal_index][col]
            pred_dicts.append({target: td})

        new_pred_dicts = []
        for row in pred_dicts:
            new_row = {}
            for key in row:
                new_row.update(row[key])
                new_row[key] = new_row['predicted_value']
            del new_row['predicted_value']
            new_pred_dicts.append(new_row)
        pred_dicts = new_pred_dicts

        columns = list(dtype_dict.keys())
        predicted_columns = target
        if not isinstance(predicted_columns, list):
            predicted_columns = [predicted_columns]
        # endregion

        original_target_values = {}
        for col in predicted_columns:
            df = df.reset_index()
            original_target_values[col + '_original'] = []
            for _index, row in df.iterrows():
                original_target_values[col + '_original'].append(row.get(col))

        # region transform ts predictions
        timeseries_settings = learn_args.get('timeseries_settings', {'is_timeseries': False})

        if timeseries_settings['is_timeseries'] is True:
            # offset forecast if have __mdb_forecast_offset > 0
            forecast_offset = any([
                row.get('__mdb_forecast_offset') is not None and row['__mdb_forecast_offset'] > 0
                for row in pred_dicts
            ])

            group_by = timeseries_settings.get('group_by', [])
            order_by_column = timeseries_settings['order_by']
            if isinstance(order_by_column, list):
                order_by_column = order_by_column[0]
            horizon = timeseries_settings['horizon']

            # region convert values to lists in case of horizon==1.
            # That needs to make processing below unified for any case.
            if horizon == 1:
                for row in pred_dicts:
                    if isinstance(row[order_by_column], list) is False:
                        row[order_by_column] = [row[order_by_column]]
                    if isinstance(row[target], list) is False:
                        row[target] = [row[target]]
                for row in explain_arr:
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        if isinstance(row[target][col], list) is False:
                            row[target][col] = [row[target][col]]
            # endregion

            if len(group_by) == 0:
                rows_by_groups = {
                    (): {
                        'rows': pred_dicts,
                        'explanations': explain_arr
                    }
                }
            else:
                groups = set()
                for row in pred_dicts:
                    groups.add(
                        tuple([row[x] for x in group_by])
                    )

                # split rows by groups
                rows_by_groups = {}
                for group in groups:
                    rows_by_groups[group] = {
                        'rows': [],
                        'explanations': []
                    }
                    for row_index, row in enumerate(pred_dicts):
                        is_wrong_group = False
                        for i, group_by_key in enumerate(group_by):
                            if row[group_by_key] != group[i]:
                                is_wrong_group = True
                                break
                        if not is_wrong_group:
                            rows_by_groups[group]['rows'].append(row)
                            rows_by_groups[group]['explanations'].append(explain_arr[row_index])

            for group, data in rows_by_groups.items():
                rows = data['rows']
                explanations = data['explanations']

                if len(rows) == 0:
                    break

                for row in rows:
                    predictions = row[target]
                    if isinstance(predictions, list) is False:
                        predictions = [predictions]

                    date_values = row[order_by_column]
                    if isinstance(date_values, list) is False:
                        date_values = [date_values]

                if pred_args.get('force_ts_infer') is True:
                    # last row contains one additional prediction (used for cases like date > '2020-10-10').
                    # Extract that prediction from there and join to previous row
                    rows[-2][order_by_column] = rows[-2][order_by_column].copy()
                    rows[-2][target] = rows[-2][target].copy()

                    rows[-2][order_by_column].append(rows[-1][order_by_column][-1])
                    rows[-2][target].append(rows[-1][target][-1])
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        explanations[-2][target][col].append(explanations[-1][target][col][-1])
                    rows.pop()
                    explanations.pop()
                    # horizon = horizon + 1

                for i in range(len(rows) - 1):
                    row_horizon = len(rows[i][target])
                    if row_horizon > 1:
                        rows[i][target] = rows[i][target][0]
                        if isinstance(rows[i][order_by_column], list):
                            rows[i][order_by_column] = rows[i][order_by_column][0]
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        if row_horizon > 1 and col in explanations[i][target]:
                            explanations[i][target][col] = explanations[i][target][col][0]

                last_row = rows.pop()
                last_explanation = explanations.pop()
                for i in range(len(last_row[target])):
                    new_row = copy.deepcopy(last_row)
                    new_row[target] = new_row[target][i]
                    if isinstance(new_row[order_by_column], list):
                        new_row[order_by_column] = new_row[order_by_column][i]
                    if '__mindsdb_row_id' in new_row and (i > 0 or forecast_offset):
                        new_row['__mindsdb_row_id'] = None

                    new_explanation = copy.deepcopy(last_explanation)
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        if col in new_explanation[target]:
                            new_explanation[target][col] = new_explanation[target][col][i]
                    if i != 0:
                        new_explanation[target]['anomaly'] = None
                        new_explanation[target]['truth'] = None

                    rows.append(new_row)
                    explanations.append(new_explanation)

            pred_dicts = []
            explanations = []
            for group, data in rows_by_groups.items():
                pred_dicts.extend(data['rows'])
                explanations.extend(data['explanations'])

            original_target_values[f'{target}_original'] = []
            for i in range(len(pred_dicts)):
                original_target_values[f'{target}_original'].append(explanations[i][target].get('truth', None))

            if dtype_dict[order_by_column] == dtype.date:
                for row in pred_dicts:
                    if isinstance(row[order_by_column], (int, float)):
                        row[order_by_column] = datetime.fromtimestamp(row[order_by_column]).date()
            elif dtype_dict[order_by_column] == dtype.datetime:
                for row in pred_dicts:
                    if isinstance(row[order_by_column], (int, float)):
                        row[order_by_column] = datetime.fromtimestamp(row[order_by_column])

            explain_arr = explanations
        # endregion

        if pred_format == 'explain':
            return explain_arr

        keys = [x for x in pred_dicts[0] if x in columns]
        min_max_keys = []
        for col in predicted_columns:
            if dtype_dict[col] in (dtype.integer, dtype.float, dtype.num_tsarray):
                min_max_keys.append(col)

        data = []
        explains = []
        keys_to_save = [*keys, '__mindsdb_row_id', 'select_data_query', 'when_data']
        for i, el in enumerate(pred_dicts):
            data.append({key: el.get(key) for key in keys_to_save})
            explains.append(explain_arr[i])

        for i, row in enumerate(data):
            cast_row_types(row, dtype_dict)

            for k in original_target_values:
                try:
                    row[k] = original_target_values[k][i]
                except Exception:
                    row[k] = None

            for column_name in columns:
                if column_name not in row:
                    row[column_name] = None

            explanation = explains[i]
            for key in predicted_columns:
                row[key + '_confidence'] = explanation[key]['confidence']
                row[key + '_explain'] = json.dumps(explanation[key], cls=NumpyJSONEncoder, ensure_ascii=False)
                if 'anomaly' in explanation[key]:
                    row[key + '_anomaly'] = explanation[key]['anomaly']
            for key in min_max_keys:
                if 'confidence_lower_bound' in explanation[key]:
                    row[key + '_min'] = explanation[key]['confidence_lower_bound']
                if 'confidence_upper_bound' in explanation[key]:
                    row[key + '_max'] = explanation[key]['confidence_upper_bound']

        return pd.DataFrame(data)

    def edit_json_ai(self, name: str, json_ai: dict):
        predictor_record = get_model_record(name=name, ml_handler_name='lightwood')
        assert predictor_record is not None

        json_ai = lightwood.JsonAI.from_dict(json_ai)
        predictor_record.code = lightwood.code_from_json_ai(json_ai)
        db.session.commit()

        json_storage = get_json_storage(
            resource_id=predictor_record.id
        )
        json_storage.set('json_ai', json_ai.to_dict())

    def code_from_json_ai(self, json_ai: dict):
        json_ai = lightwood.JsonAI.from_dict(json_ai)
        code = lightwood.code_from_json_ai(json_ai)
        return code

    def edit_code(self, name: str, code: str):
        """Edit an existing predictor's code"""
        if self.config.get('cloud', False):
            raise Exception('Code editing prohibited on cloud')

        predictor_record = get_model_record(name=name, ml_handler_name='lightwood')
        assert predictor_record is not None

        lightwood.predictor_from_code(code)
        predictor_record.code = code
        db.session.commit()

        json_storage = get_json_storage(
            resource_id=predictor_record.id
        )
        json_storage.delete('json_ai')

    def _get_features_info(self):
        ai_info = self.model_storage.json_get('json_ai')
        if ai_info == {}:
            raise Exception("predictor doesn't contain enough data to generate 'feature' attribute.")
        data = []
        dtype_dict = ai_info["dtype_dict"]
        for column in dtype_dict:
            c_data = []
            c_data.append(column)
            c_data.append(dtype_dict[column])
            c_data.append(ai_info["encoders"][column]["module"])
            if ai_info["encoders"][column]["args"].get("is_target", "False") == "True":
                c_data.append("target")
            else:
                c_data.append("feature")
            data.append(c_data)

        return pd.DataFrame(data, columns=['column', 'type', 'encoder', 'role'])

    def _get_model_info(self):
        json_ai = self.model_storage.json_get('json_ai')
        model_info = self.model_storage.get_info()
        model_data = model_info['data']

        accuracy_functions = json_ai.get('accuracy_functions')
        if accuracy_functions:
            accuracy_functions = str(accuracy_functions)

        models_data = model_data.get("submodel_data", [])
        if models_data == []:
            raise Exception("predictor doesn't contain enough data to generate 'model' attribute")
        data = []

        for model in models_data:
            m_data = []
            m_data.append(model["name"])
            m_data.append(model["accuracy"])
            m_data.append(model.get("training_time", "unknown"))
            m_data.append(1 if model["is_best"] else 0)
            m_data.append(accuracy_functions)
            data.append(m_data)

        return pd.DataFrame(data, columns=['name', 'performance', 'training_time', 'selected', 'accuracy_functions'])

    def _get_ensemble_data(self):
        ai_info = self.model_storage.json_get('json_ai')
        if ai_info == {}:
            raise Exception("predictor doesn't contain enough data to generate 'ensamble' attribute. Please wait until predictor is complete.")
        ai_info_str = json.dumps(ai_info, indent=2)

        return pd.DataFrame([[ai_info_str]], columns=['ensemble'])

    def _get_progress_data(self):
        progress_info = self.model_storage.training_state_get()
        return pd.DataFrame([progress_info], columns=["current", "total", "name"])

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        if attribute == 'info':

            model_description = {}

            model_info = self.model_storage.get_info()
            model_data = model_info['data']
            to_predict = model_info['to_predict'][0]

            if model_data.get('accuracies', None) is not None:
                if len(model_data['accuracies']) > 0:
                    model_data['accuracy'] = float(np.mean(list(model_data['accuracies'].values())))

            model_columns = self.model_storage.columns_get()

            model_description['accuracies'] = model_data['accuracies']
            model_description['column_importances'] = model_data['column_importances']
            model_description['outputs'] = [to_predict]
            model_description['inputs'] = [col for col in model_columns if col not in model_description['outputs']]

            return pd.DataFrame([model_description])

        elif attribute == "features":
            return self._get_features_info()

        elif attribute == "model":
            return self._get_model_info()

        elif attribute == "jsonai":
            return self._get_ensemble_data()

        elif attribute == "progress":
            # todo remove?
            return self._get_progress_data()

        else:
            tables = ['info', 'features', 'model', 'jsonai']
            return pd.DataFrame(tables, columns=['tables'])





from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers

from .is_master import responder as ismaster_responder


class Responce(Responder):
    when = {'ismaster': helpers.is_true}

    result = ismaster_responder.result


responder = Responce()



import os
import time
import tempfile
import threading
from pathlib import Path
from typing import Optional

import psutil
from appdirs import user_data_dir


def create_directory(path):
    path = Path(path)
    path.mkdir(mode=0o777, exist_ok=True, parents=True)


def get_root_path():
    mindsdb_path = user_data_dir('mindsdb', 'mindsdb')
    return os.path.join(mindsdb_path, 'var/')


def get_or_create_data_dir():
    data_dir = user_data_dir('mindsdb', 'mindsdb')
    mindsdb_data_dir = os.path.join(data_dir, 'var/')

    if os.path.exists(mindsdb_data_dir) is False:
        create_directory(mindsdb_data_dir)

    try:
        assert os.path.exists(mindsdb_data_dir)
        assert os.access(mindsdb_data_dir, os.W_OK) is True
    except Exception:
        raise Exception('MindsDB storage directory does not exist and could not be created')

    return mindsdb_data_dir


def create_dirs_recursive(path):
    if isinstance(path, dict):
        for p in path.values():
            create_dirs_recursive(p)
    elif isinstance(path, str):
        create_directory(path)
    else:
        raise ValueError(f'Wrong path: {path}')


def _get_process_mark_id(unified: bool = False) -> str:
    ''' Creates a text that can be used to identify process+thread
        Args:
            unified: bool, if True then result will be same for same process+thread
        Returns:
            mark of process+thread
    '''
    mark = f'{os.getpid()}-{threading.get_native_id()}'
    if unified is True:
        return mark
    return f"{mark}-{str(time.time()).replace('.', '')}"


def create_process_mark(folder='learn'):
    mark = None
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath(f'mindsdb/processes/{folder}/')
        p.mkdir(parents=True, exist_ok=True)
        mark = _get_process_mark_id()
        p.joinpath(mark).touch()
    return mark


def delete_process_mark(folder: str = 'learn', mark: Optional[str] = None):
    if mark is None:
        mark = _get_process_mark_id()
    if os.name == 'posix':
        p = (
            Path(tempfile.gettempdir())
            .joinpath(f'mindsdb/processes/{folder}/')
            .joinpath(mark)
        )
        if p.exists():
            p.unlink()


def clean_process_marks():
    """ delete all existing processes marks
    """
    if os.name != 'posix':
        return

    p = Path(tempfile.gettempdir()).joinpath('mindsdb/processes/')
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            return
        for file in path.iterdir():
            file.unlink()


def clean_unlinked_process_marks():
    """ delete marks that does not have corresponded processes/threads
    """
    if os.name != 'posix':
        return

    p = Path(tempfile.gettempdir()).joinpath('mindsdb/processes/')
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            return
        for file in path.iterdir():
            parts = file.name.split('-')
            process_id = int(parts[0])
            thread_id = int(parts[1])

            try:
                process = psutil.Process(process_id)
                if process.status() in (psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD):
                    raise psutil.NoSuchProcess(process_id)

                threads = process.threads()
                try:
                    next(t for t in threads if t.id == thread_id)
                except StopIteration:
                    from mindsdb.utilities.log import get_log
                    get_log('main').warning(
                        f'We have mark for process/thread {process_id}/{thread_id} but it does not exists'
                    )
                    file.unlink()

            except psutil.NoSuchProcess:
                from mindsdb.utilities.log import get_log
                get_log('main').warning(
                    f'We have mark for process/thread {process_id}/{thread_id} but it does not exists'
                )
                file.unlink()



from ..mysql_handler import Handler as MySQLHandler, connection_args, connection_args_example


class PlanetScaleHandler(MySQLHandler):
    """
    This handler handles connection and execution of queries against PlanetScale.
    """
    name = 'planet_scale'



from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.utilities.date_utils import interval_str_to_duration_ms, utc_date_str_to_timestamp_ms
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser import ast

from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List

import pandas as pd
import time


class BinanceAggregatedTradesTable(APITable):

    # Default 1m intervals in aggregate data.
    DEFAULT_AGGREGATE_TRADE_INTERVAL = '1m'
    DEFAULT_AGGREGATE_TRADE_LIMIT = 1000
    # Binance Spot client has connection pool size of 10.
    MAX_THREAD_POOL_WORKERS = 10

    def _get_batch_klines(self, executor: ThreadPoolExecutor, total_results: int, params: Dict) -> pd.DataFrame:
        """Gets aggregate trade data in batches and combines the results together.
        
        Returns all results as a pandas DataFrame.

        Args:
            executor (ThreadPoolExecutor): Executor to use when mapping API calls as tasks.
            total_results (int): Total number of results to fetch.
            params (Dict): Overall request params to be split into batches.
        """
        interval_duration_ms = interval_str_to_duration_ms(params['interval'])
        if 'end_time' not in params:
            # Default to get all klines before the current time.
            overall_end_ms = int(time.time() * 1000)
        else:
            overall_end_ms = params['end_time']

        if 'start_time' not in params:
            total_duration_ms = interval_duration_ms * total_results
            # Infer start time based on the interval and how many klines we need to fetch.
            overall_start_ms = overall_end_ms - total_duration_ms
        else:
            overall_start_ms = params['start_time']

        next_params = params.copy()
        next_params['start_time'] = overall_start_ms
        duration_per_api_call_ms = interval_duration_ms * BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT
        next_params['end_time'] = min(overall_end_ms, overall_start_ms + duration_per_api_call_ms)
        all_params = [next_params.copy()]
        results_so_far = BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT
        while next_params['end_time'] < overall_end_ms and results_so_far < total_results:
            next_params['limit'] = min(
                BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT,
                total_results - results_so_far
            )
            next_params['start_time'] = next_params['end_time']
            next_params['end_time'] = min(overall_end_ms, next_params['start_time'] + duration_per_api_call_ms)
            all_params.append(next_params.copy())
            results_so_far += next_params['limit']

        aggregated_trade_subdatas = list(executor.map(lambda p: self.handler.call_binance_api(method_name='klines', params=p), all_params))
        if not aggregated_trade_subdatas:
            return pd.DataFrame([])

        aggregated_trade_data = aggregated_trade_subdatas[0]
        for aggregated_trade_subdata in aggregated_trade_subdatas[1:]:
            aggregated_trade_data = pd.concat([aggregated_trade_data, aggregated_trade_subdata])
        return aggregated_trade_data

    def _get_kline_params_from_conditions(self, conditions: List) -> Dict:
        """Gets aggregate trade data API params from SQL WHERE conditions.
        
        Returns params to use for Binance API call to klines.

        Args:
            conditions (List): List of individual SQL WHERE conditions.
        """
        params = {
            'interval': BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_INTERVAL,
            'limit': BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT
        }
        for op, arg1, arg2 in conditions:
            if arg1 == 'interval':
                if op != '=':
                    raise NotImplementedError
                params['interval'] = arg2

            elif arg1 == 'symbol':
                if op != '=':
                    raise NotImplementedError
                params['symbol'] = arg2
        interval_duration_ms = interval_str_to_duration_ms(params['interval'])

        for op, arg1, arg2 in conditions:
            if arg1 == 'open_time':
                utc_timestamp_ms = utc_date_str_to_timestamp_ms(arg2)
                if op == '>':
                    params['start_time'] = utc_timestamp_ms
                else:
                    raise NotImplementedError
                continue
            elif arg1 == 'close_time':
                utc_timestamp_ms = utc_date_str_to_timestamp_ms(arg2)
                if op == '<':
                    params['end_time'] = utc_timestamp_ms - interval_duration_ms
                else:
                    raise NotImplementedError

        return params

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the Binance API and returns it as a pandas DataFrame.
        
        Returns dataframe representing the Binance API results.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)
        params = self._get_kline_params_from_conditions(conditions)

        total_results = params['limit']
        if query.limit:
            total_results = query.limit.value
            params['limit'] = min(BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT, query.limit.value)

        if total_results > BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT:
            # Max 1000 klines per API call so we need to combine multiple API calls.
            with ThreadPoolExecutor(max_workers=BinanceAggregatedTradesTable.MAX_THREAD_POOL_WORKERS) as executor:
                aggregated_trades_data = self._get_batch_klines(executor, total_results, params)

        else:
            aggregated_trades_data = self.handler.call_binance_api(
                method_name='klines',
                params=params
            )

        # Only return the columns we need to.
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(aggregated_trades_data) == 0:
            aggregated_trades_data = pd.DataFrame([], columns=columns)
        else:
            # Remove columns not part of select.
            aggregated_trades_data.columns = self.get_columns()
            for col in set(aggregated_trades_data.columns).difference(set(columns)):
                aggregated_trades_data = aggregated_trades_data.drop(col, axis=1)

        return aggregated_trades_data

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'symbol',
            'open_time',
            'open_price',
            'high_price',
            'low_price',
            'close_price',
            'volume',
            'close_time',
            'quote_asset_volume',
            'number_of_trades',
            'taker_buy_base_asset_volume',
            'taker_buy_quote_asset_volume'
        ]



import pyodbc

from collections import OrderedDict
import pandas as pd
from mindsdb_sql import parse_sql

from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class EmpressHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Empress Embedded statements.
    """

    name = 'empress'

    def __init__(self, name: str, **kwargs):
        """
        Initializes a new instance of the Empress Embedded handler.

        Args:
            name (str): The name of the database.
            connection_data (dict): parameters for connecting to the database
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'empress'
        self.connection_args = kwargs.get('connection_data')
        self.database = self.connection_args.get('database')
        self.server = self.connection_args.get('server')
        self.user = self.connection_args.get('user')
        self.password = self.connection_args.get('password')
        self.host = self.connection_args.get('host')
        self.port = self.connection_args.get('port', 6322)
        self.connection = None
        self.is_connected = False

    def __del__(self):
        """
        Destructor for the Empress Embedded class.
        """
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Establishes a connection to the Empress Embedded server.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected:
            return self.connection

        conn_str = f"DRIVER={{Empress ODBC Interface [Default]}};Server={self.server};Port={self.port};UID={self.user};PWD={self.password};Database={self.database};"
        self.connection = pyodbc.connect(conn_str)
        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Empress Embedded, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def disconnect(self):
        """
        Closes the connection to the Empress Embedded server.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def native_query(self, query: str) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): SQL query to execute.
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame.from_records(
                            result,
                            columns=[x[0] for x in cursor.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.connection_args["database"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender('sqlite')

        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Gets a list of table names in the database.

        Returns:
            list: A list of table names in the database.
        """
        connection = self.connect()
        cursor = connection.cursor()
        # Execute query to get all table names
        cursor.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema='public' AND table_type='BASE TABLE'")

        table_names = [x[0] for x in cursor.fetchall()]

        # Create dataframe with table names
        df = pd.DataFrame(table_names, columns=['table_name', 'data_type'])

        # Create response object
        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Gets a list of column names in the specified table.

        Args:
            table_name (str): The name of the table to get column names from.

        Returns:
            list: A list of column names in the specified table.
        """
        conn = self.connect()
        cursor = conn.cursor()
        cursor.execute("SELECT column_name FROM information_schema.columns WHERE table_name='{}'".format(table_name))
        results = cursor.fetchall()

        # construct a pandas dataframe from the query results
        df = pd.DataFrame(
            results,
            columns=['column_name', 'data_type']
        )

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Empress Embedded server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect to Empress Embedded server'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Empress Embedded server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Empress Embedded server.'
    },
    server={
        'type': ARG_TYPE.STR,
        'description': 'The server name used to authenticate with the Empress Embedded server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name to connect Empress Embedded server'
    },

)

connection_args_example = OrderedDict(
    host='localhost',
    port=6322,
    user='admin',
    password='password',
    server='test',
    database='test_db'
)
